import gc
import json
import os
import pathlib

from typing import List

import qtawesome as qta

from pathlib import Path

from PyQt6.QtCore import pyqtSlot, QObject, QUrl
from PyQt6.QtWebChannel import QWebChannel
from PyQt6.QtWebEngineCore import QWebEngineSettings
from PyQt6.QtWebEngineWidgets import QWebEngineView
from PyQt6.QtWidgets import QVBoxLayout, QLabel, QWidget, QFrame, QHBoxLayout, QSizePolicy

import folium.features

from otlmow_model.OtlmowModel.BaseClasses.OTLObject import OTLObject
from otlmow_visuals.PyVisWrapper import PyVisWrapper

from otlmow_gui.Domain import global_vars
from otlmow_gui.Domain.ProgramFileStructure import ProgramFileStructure
from otlmow_gui.Domain.logger.OTLLogger import OTLLogger
from otlmow_gui.Domain.step_domain.RelationChangeDomain import RelationChangeDomain
from otlmow_gui.GUI.dialog_windows.LoadingImageWindow import add_loading_screen
from otlmow_gui.GUI.screens.Map_elements.MapHelper import MapHelper
from otlmow_gui.GUI.screens.general_elements.ButtonWidget import ButtonWidget
from otlmow_gui.GUI.screens.Screen import Screen
from otlmow_gui.exception_handler.ExceptionHandlers import create_task_reraise_exception

ROOT_DIR = Path(__file__).parent

IMG_DIR = ProgramFileStructure.get_dynamic_library_path('img')

HTML_DIR = Path.home() / 'OTLWizardProjects' / 'img' / 'html'


class WebBridge(QObject):
    """Bridge between JavaScript and Python using QWebChannel."""

    def __init__(self,folium_map: folium.Map):
        super().__init__()
        self.folium_map = folium_map


    @pyqtSlot(str)
    def receive_coordinates(self, message):
        """Receives click event data from JavaScript."""
        data = json.loads(message)  # Convert string back to dict
        lat = float(data["lat"])
        lng =float(data["lng"])


        # bol_icon = folium.features.CustomIcon(str(ROOT_DIR.parent.parent / "img"/"bol.png"),
        #                                      icon_size=(30, 30))
        # folium.Marker([lat, lng],
        #               # icon=bol_icon,
        #               # popup='red',
        #               ).add_to(self.folium_map)

    @pyqtSlot(str)
    def  receive_selection_id(self, message):
        """Receives click event data from JavaScript."""
        data = json.loads(message)  # Convert string back to dict
        asset_id = str(data["id"])

        RelationChangeDomain.set_selected_object_from_map(asset_id)
        RelationChangeDomain.update_frontend()

class MapScreen(Screen):

    object_count_limit = 300

    def __init__(self, _, parent_screen):
        super().__init__()

        if not HTML_DIR.exists():
            os.makedirs(HTML_DIR,exist_ok=True)
        self.map = None
        self.prev_selected_asset_id = None
        self.parent_screen = parent_screen

        self.frame_layout_legend = None
        self.refresh_needed_label = QLabel()
        self.container_insert_data_screen = QVBoxLayout()
        self._ = _
        self.webView = QWebEngineView()
        self.too_many_objects_message = QLabel()
        self.color_label_title = QLabel()
        self.init_ui()

        self.channel = QWebChannel()
        self.img_qurl = QUrl(pathlib.Path.home().drive + str(
            (IMG_DIR / "bol.png").absolute()).replace("\\","/"))

        # self.relation_change_screen_object_list_content_dict = self.load_assets()

        # map_path, self.map , self.map_id = MapHelper.create_html_map(self.relation_change_screen_object_list_content_dict, ROOT_DIR, HTML_DIR,self.prev_selected_asset_id)

        self.map = MapHelper.create_folium_map()
        map_path = MapHelper.get_map_html_save_path(HTML_DIR,ROOT_DIR)
        self.map.save(map_path)
        self.map_id = self.map.get_name()

        self.web_bridge = WebBridge(self.map)
        self.channel.registerObject("webBridge", self.web_bridge)

        # with open(map_path, 'r', encoding='utf-8') as f:
        #     html_content = f.read()
        # self.webView.setHtml(html_content)
        map_QUrl = QUrl.fromLocalFile(map_path.replace("\\", "/"))
        self.webView.setUrl(map_QUrl)
        OTLLogger.logger.debug(map_QUrl.toString())

        self.webView.page().setWebChannel(self.channel)


        # MapHelper.zoom_to_assets(web_view=self.webView,map_id= self.map_id,prev_selected_asset_id=self.prev_selected_asset_id)

        self.setLayout(self.container_insert_data_screen)



    def init_ui(self):
        self.container_insert_data_screen.addWidget(self.create_html_container())

    def create_html_container(self):
        window = QWidget()
        window.setProperty('class', 'background-box')
        
        window_layout = QVBoxLayout()
        window_layout.setContentsMargins(0,0,0,0)

        self.webView.settings().setAttribute(QWebEngineSettings.WebAttribute.ShowScrollBars, False)
        self.webView.setContentsMargins(0, 0, 0, 0)
        self.webView.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)
        self.webView.settings().setAttribute(
            QWebEngineSettings.WebAttribute.LocalContentCanAccessFileUrls, True)
        self.webView.settings().setAttribute(
            QWebEngineSettings.WebAttribute.LocalContentCanAccessRemoteUrls, True)
        self.too_many_objects_message.setVisible(False)

        self.color_label_title.setText(self._("relations legend") + ":")
        
        window_layout.addWidget(self.create_button_container())
        window_layout.addWidget(self.webView, 2)
        window_layout.addWidget(self.too_many_objects_message,2)
        
        window.setLayout(window_layout)
        return window


    def create_color_legend(self):

        frame = QFrame()
        self.frame_layout_legend = QHBoxLayout()
        self.fill_frame_layout_legend()
        frame.setLayout(self.frame_layout_legend)
        frame.setHidden(True)
        return frame

    def create_button_container(self):
        frame = QFrame()
        frame_layout = QHBoxLayout()
        
        refresh_btn = ButtonWidget()
        refresh_btn.setIcon(qta.icon('mdi.refresh', color='white'))
        refresh_btn.setProperty('class', 'primary-button')
        refresh_btn.clicked.connect(lambda: self.start_async_reload())

        self.refresh_needed_label.setText(self._("The visualisation is outdated, refresh to see new changes"))
        self.refresh_needed_label.setStyleSheet('color:#DD1111;')

        frame_layout.addWidget(refresh_btn)
        frame_layout.addWidget(self.refresh_needed_label)
        frame_layout.addStretch()
        frame_layout.setContentsMargins(0,0,0,0)
        
        frame.setLayout(frame_layout)
        return frame

    def reset_ui(self, _):
        super().reset_ui(_)

        self._ = _
        self.color_label_title.setText(self._("relations legend") + ":")

    def start_async_reload(self):
        create_task_reraise_exception(self.reload_html())

    @add_loading_screen
    async def reload_html(self):
        OTLLogger.logger.debug(
            f"Executing MapScreen.reload_html() for project {global_vars.current_project.eigen_referentie}",
            extra={"timing_ref": f"reload_html_{global_vars.current_project.eigen_referentie}"})

        # throw away old data before loading the new
        self.relation_change_screen_object_list_content_dict = {}
        gc.collect()

        self.relation_change_screen_object_list_content_dict = self.load_assets()

        map_path, self.map , self.map_id = await MapHelper.create_html_map(self.relation_change_screen_object_list_content_dict,
                                                                     IMG_DIR, HTML_DIR,self.prev_selected_asset_id)
        self.web_bridge.folium_map = self.map
        # self.webView.setHtml(open(map_path).read())
        map_QUrl = QUrl.fromLocalFile(map_path.replace("\\", "/"))
        self.webView.setUrl(map_QUrl)
        OTLLogger.logger.debug(map_QUrl.toString())

        # self.webView.page().setWebChannel(self.channel)

        first_coordinate = None


        object_count = len(self.relation_change_screen_object_list_content_dict)
        OTLLogger.logger.debug(
            f"Executing MapScreen.reload_html() for project {global_vars.current_project.eigen_referentie} ({object_count} objects)",
            extra={"timing_ref": f"reload_html_{global_vars.current_project.eigen_referentie}"})


    def load_assets(self) -> dict:
        assets = RelationChangeDomain.get_current_relation_change_screen_object_list_content_dict()
        self.check_if_refresh_message_is_needed()
        return  assets

    def create_html(self, objects_in_memory:List[OTLObject]):
        object_count = len(objects_in_memory)
        if object_count > MapScreen.object_count_limit:
            self.webView.setVisible(False)
            if not self.too_many_objects_message.isVisible():
                self.too_many_objects_message.setVisible(True)

            translation = self._("There are too many assets and relations to create a visualisation.\nmaximum: {0}\ncurrent: {1}")
            self.too_many_objects_message.setText(translation.format(
                MapScreen.object_count_limit,
                object_count))

        else:
            self.too_many_objects_message.setVisible(False)
            if not self.webView.isVisible():
                self.webView.setVisible(True)
            html_loc = HTML_DIR / "visuals.html"
            previous_cwd = os.getcwd()
            os.chdir(Path.home() / 'OTLWizardProjects')
            PyVisWrapper().show(list_of_objects=objects_in_memory,
                                html_path=Path(html_loc), launch_html=False)
            os.chdir(previous_cwd)
            resources =  QUrl().fromLocalFile((os.path.dirname(os.path.realpath(__file__))))
            map_QUrl = QUrl.fromLocalFile(str(html_loc).replace("\\", "/"))
            self.webView.setUrl(map_QUrl)
            OTLLogger.logger.debug(map_QUrl.toString())
            # self.webView.setHtml(open(html_loc).read(),resources)

    def opened(self):
        self.check_if_refresh_message_is_needed()

    def check_if_refresh_message_is_needed(self):
        self.refresh_needed_label.setHidden(RelationChangeDomain.is_visualisation_uptodate())

    def activate_highlight_layer_by_id(self, asset_id:str):
        self.prev_selected_asset_id = asset_id
        MapHelper.activate_highlight_layer_by_id(asset_id,self.webView,self.map_id)

    def closeEvent(self, a0):
        if self.parent_screen:
            self.parent_screen.map_window = None
        super().closeEvent(a0)




from otlmow_converter.Exceptions.FailedToImportFileError import FailedToImportFileError
from otlmow_converter.Exceptions.InvalidColumnNamesInExcelTabError import \
    InvalidColumnNamesInExcelTabError
from otlmow_converter.Exceptions.NoTypeUriInTableError import NoTypeUriInTableError
from otlmow_converter.Exceptions.TypeUriNotInFirstRowError import TypeUriNotInFirstRowError

from otlmow_gui.Exceptions.NoIdentificatorError import NoIdentificatorError
from otlmow_gui.Exceptions.RelationHasInvalidTypeUriForSourceAndTarget import \
    RelationHasInvalidTypeUriForSourceAndTarget
from otlmow_gui.Exceptions.RelationHasNonExistingTypeUriForSourceOrTarget import \
    RelationHasNonExistingTypeUriForSourceOrTarget
from otlmow_gui.GUI.translation.GlobalTranslate import GlobalTranslate


class ValidationErrorReportTranslations:



    @classmethod
    def translate_exception(cls, doc_name, exception):
        if str(exception) == "argument of type 'NoneType' is not iterable":
            error_text = GlobalTranslate._(
                "{doc_name}: Data nodig in een datasheet om objecten in te laden.\n").format(
                doc_name=doc_name)
        elif issubclass(type(exception), NoTypeUriInTableError):
            error_text = GlobalTranslate._(
                "{doc_name}: No type uri in {tab}\n").format(
                doc_name=doc_name, tab=str(exception.tab))
        elif issubclass(type(exception), InvalidColumnNamesInExcelTabError):
            error_text = GlobalTranslate._(
                "{doc_name}: invalid columns in {tab}, bad columns are {bad_columns} \n").format(
                doc_name=doc_name, tab=exception.tab, bad_columns=str(exception.bad_columns))
        elif issubclass(type(exception), TypeUriNotInFirstRowError):
            error_text = GlobalTranslate._(
                "{doc_name}: type uri not in first row of {tab}\n").format(
                doc_name=doc_name, tab=str(exception.tab))
        elif issubclass(type(exception),
                        FailedToImportFileError):  # as of otlmow_converter==1.4 never instantiated
            error_text = GlobalTranslate._(f'{doc_name}: {exception} \n')
        elif issubclass(type(exception), NoIdentificatorError):
            error_text = GlobalTranslate._(
                "{doc_name}: There are assets without an assetId.identificator in "
                "worksheet {tab}\n").format(doc_name=doc_name, tab=str(exception.tab))
        elif issubclass(type(exception), RelationHasInvalidTypeUriForSourceAndTarget):
            error_text = GlobalTranslate._(
                "{doc_name}:\n" +
                "Relation of type: \"{type_uri}\"\n" +
                "with assetId.identificator: \"{identificator}\"\n" +
                "This relation cannot be made between the typeURI's.\n" +
                "{wrong_field}= \"{wrong_value}\"\n" +
                "{wrong_field2}= \"{wrong_value2}\"\nin tab {tab}\n").format(
                doc_name=doc_name,
                type_uri=exception.relation_type_uri,
                identificator=exception.relation_identificator,
                wrong_field=exception.wrong_field,
                wrong_value=exception.wrong_value,
                wrong_field2=exception.wrong_field2,
                wrong_value2=exception.wrong_value2,
                tab=exception.tab)
        elif issubclass(type(exception), RelationHasNonExistingTypeUriForSourceOrTarget):
            error_text = GlobalTranslate._(
                "{doc_name}:\n"
                "Relation of type: \"{type_uri}\"\n"
                "with assetId.identificator: \"{identificator}\",\n"
                "has the non-existing TypeURI value: \"{wrong_value}\"\n"
                "for field \"{wrong_field}\".\nin tab {tab}\n").format(
                doc_name=doc_name,
                type_uri=exception.relation_type_uri,
                identificator=exception.relation_identificator,
                wrong_field=exception.wrong_field,
                wrong_value=exception.wrong_value,
                tab=exception.tab)

        else:
            error_text = GlobalTranslate._(f'{doc_name}: {exception}\n')
        return error_text


from pathlib import Path
from typing import Callable

from PyQt6.QtWidgets import QDialog, QHBoxLayout, QPushButton, QStackedWidget

from otlmow_gui.Domain.enums import Language
from otlmow_gui.Domain.Settings import Settings
from otlmow_gui.GUI.dialog_windows.NotificationWindow import NotificationWindow

ROOT_DIR = Path(__file__).parent.parent

LANG_DIR = ROOT_DIR.parent / 'locale/'


class LanguageWindow(QDialog):

    def __init__(self, language_settings: Callable, main_window: QStackedWidget):
        super().__init__()
        self._ = language_settings
        self.setModal(True)
        self.setWindowTitle(self._("change_language_title"))

        layout = QHBoxLayout()

        button_ned = QPushButton(self._("language_option_dutch"))
        button_eng = QPushButton(self._("language_option_english"))
        button_ned.clicked.connect(lambda: self.change_language(lang= Language.DUTCH, main_window= main_window))
        button_eng.clicked.connect(lambda: self.change_language(lang= Language.ENGLISH, main_window= main_window))
        layout.addWidget(button_ned)
        layout.addWidget(button_eng)
        self.setLayout(layout)

    def change_language(self, lang: Language, main_window: QStackedWidget) -> None:
        """
        Changes the application's language and updates the user interface accordingly.

        This method updates the language setting in the application's settings file, retrieves the
        corresponding translation Callable, and resets the main window's UI to reflect the new
        language. Finally, it closes the language selection window.

        :param lang: The new language to be set, represented as a Language enum.
        :type lang: Language
        :param main_window: The main window instance to be updated.
        :type main_window: QStackedWidget

        :return: None
        """
        Settings.change_language_on_settings_file(lang)
        restart_request_dialog = NotificationWindow(message=self._("Please restart the application to apply the changes"),
                                                    title=self._("Restart required"))
        restart_request_dialog.exec()
        # settings = Settings.get_or_create_settings_file()
        # self._ = GlobalTranslate(settings, LANG_DIR).get_all()
        # main_window.reset_ui(self._)
        self.close()



import asyncio
import os
import platform
import traceback
from pathlib import Path
from typing import List

from otlmow_converter.Exceptions.ExceptionsGroup import ExceptionsGroup
from otlmow_model.OtlmowModel.Classes.Agent import Agent
from otlmow_modelbuilder.SQLDataClasses.OSLOClass import OSLOClass
from otlmow_template.SubsetTemplateCreator import SubsetTemplateCreator
from otlmow_gui.Domain import global_vars
from otlmow_gui.Domain.util.Helpers import Helpers
from otlmow_gui.Domain.util.SDFHandler import SDFHandler
from otlmow_gui.Domain.logger.OTLLogger import OTLLogger
from otlmow_gui.Domain.network.Updater import Updater
from otlmow_gui.Exceptions.FDOToolboxNotInstalledError import FDOToolboxNotInstalledError
from otlmow_gui.GUI.dialog_windows.LoadingImageWindow import add_loading_screen, LoadingImageWindow
from otlmow_gui.GUI.dialog_windows.NotificationWindow import NotificationWindow
from otlmow_gui.GUI.dialog_windows.SuggestUpdateWindow import SuggestUpdateWindow
from otlmow_gui.GUI.screens.screen_interface.TemplateScreenInterface import TemplateScreenInterface
from otlmow_gui.GUI.translation.GlobalTranslate import GlobalTranslate
from otlmow_gui.exception_handler.ExceptionHandlers import create_task_reraise_exception


class TemplateDomain:

    classes: list[OSLOClass] = []
    selected_classes_indexes: list[int] = []
    all_classes_selected: bool = True
    has_a_class_with_deprecated_attributes: bool = False
    has_agent:bool = False

    @classmethod
    def check_for_no_deprecated_present(cls) -> bool:
        return all(len(value.deprecated_version) == 0 for value in cls.classes)

    @classmethod
    def check_for_agent(cls) -> bool:
        return any(Agent.typeURI == value.objectUri for value in cls.classes)

    @classmethod
    @add_loading_screen
    async def create_template(cls, subset_path: Path, document_path: Path, selected_classes_typeURI_list: List, generate_choice_list: bool,
                              add_geo_artefact: bool, add_attribute_info: bool, highlight_deprecated_attributes: bool,
                              amount_of_examples: int, model_directory: Path = None) -> None:
        OTLLogger.logger.debug("Creating template", extra={"timing_ref": f"create_template_{document_path.name}"})
        try:
            template_creator = SubsetTemplateCreator()

            if not selected_classes_typeURI_list:
                # if selected_class_dir is empty fill it with all classes in the domain
                selected_classes_typeURI_list = [oslo_class.objectUri for oslo_class in cls.classes]

            if selected_classes_typeURI_list and 'http://purl.org/dc/terms/Agent' in selected_classes_typeURI_list:
                # filter out agent
                selected_classes_typeURI_list.remove('http://purl.org/dc/terms/Agent')

            if document_path.suffix == ".sdf":
                await SDFHandler.create_filtered_SDF_from_subset(
                    subset_path=subset_path,
                    sdf_path=document_path,
                    selected_classes_typeURI_list=selected_classes_typeURI_list,
                    model_directory=model_directory)

            else:
                await template_creator.generate_template_from_subset_async(
                    subset_path=subset_path, template_file_path=document_path,
                    class_uris_filter=selected_classes_typeURI_list,
                    generate_choice_list=generate_choice_list,
                    add_geometry=add_geo_artefact, add_attribute_info=add_attribute_info,
                    add_deprecated=highlight_deprecated_attributes,
                    dummy_data_rows=amount_of_examples,
                    model_directory=model_directory,
                    abbreviate_excel_sheettitles=True)
        except FDOToolboxNotInstalledError as e:
            pass
            # document_path_str = str(document_path)
            # OTLLogger.logger.debug(f"Permission to file was denied: {document_path_str}")
            # NotificationWindow(GlobalTranslate._("permission_to_file_was_denied_likely_due_to_the_file_being_open_in_excel") + ":\n" + document_path_str, title=GlobalTranslate._("permission_denied"))
        except PermissionError as e:
            document_path_str = str(document_path)
            OTLLogger.logger.debug(f"Permission to file was denied: {document_path_str}")
            NotificationWindow(GlobalTranslate._(
                "permission_to_file_was_denied_likely_due_to_the_file_being_open_in_excel") + ":\n" + document_path_str,
                               title=GlobalTranslate._("permission_denied"))

        except ExceptionsGroup as e:
            OTLLogger.logger.debug("Error while creating template")
            OTLLogger.logger.error(e)
            OTLLogger.logger.error("".join(traceback.format_exception(e)))
            for ex in e.exceptions:
                OTLLogger.logger.error(ex)
                OTLLogger.logger.error("".join(traceback.format_exception(ex)))

            LoadingImageWindow.attempt_destoy_loading_screen(ref="crash")
            raise e
        except Exception as e:
            OTLLogger.logger.debug("Error while creating template")
            OTLLogger.logger.error(e)
            OTLLogger.logger.error("".join(traceback.format_exception(e)))
            LoadingImageWindow.attempt_destoy_loading_screen(ref="crash")
            raise e
        OTLLogger.logger.debug("Creating template", extra={"timing_ref": f"create_template_{document_path.name}"})

    @classmethod
    def update_subset_information(cls, screen):
        screen.reset_ui( screen._)

    @classmethod
    def init_static(cls):
        if global_vars.current_project:
            create_task_reraise_exception(cls.fill_list())

    @classmethod
    async def fill_list(cls):
        cls.classes.clear()
        cls.has_a_class_with_deprecated_attributes = False
        cls.get_screen().set_gui_list_to_loading_state()
        OTLLogger.logger.debug("Load OTL classes from Subset", extra={"timing_ref": f"class_from_subset_{global_vars.current_project.eigen_referentie}"})
        try:
            await asyncio.sleep(0)  # Give the UI thread the chance to switch the screen to
                                    # TemplateScreen

            modelbuilder = global_vars.current_project.get_model_builder()
            subset_otl_version = global_vars.current_project.get_otl_version()

            otl_model_version = await Updater.async_get_local_otl_model_library_version()
            if not Helpers.is_version_equal_or_higher(otl_model_version, subset_otl_version):
                OTLLogger.logger.info("otlmow-model version is outdated")
                SuggestUpdateWindow(language_settings=GlobalTranslate._,
                                    local_version=Updater.local_version,
                                    new_version=Updater.master_version,
                                    otl_model_out_of_date=True)
            else:
                OTLLogger.logger.info("otlmow-model version is high enough")

            cls.classes = modelbuilder.filter_relations_and_abstract()

            cls.has_a_class_with_deprecated_attributes = TemplateDomain.check_for_no_deprecated_present()
            cls.has_agent = TemplateDomain.check_for_agent()
            cls.select_all_classes()


        except FileNotFoundError as e:
            #TODO: give proper feedback to user if the subset file is not found
            cls.get_screen().set_gui_list_to_no_classes_found()
        OTLLogger.logger.debug("Load OTL classes from Subset", extra={"timing_ref": f"class_from_subset_{global_vars.current_project.eigen_referentie}"})

        cls.update_frontend()
        cls.get_screen().reset_ui(GlobalTranslate._)

    @classmethod
    def select_all_classes(cls):
        total_class_count = cls.get_total_amount_of_classes()
        if cls.has_agent:
            cls.selected_classes_indexes = list(range(total_class_count + 1))
        else:
            cls.selected_classes_indexes = list(range(total_class_count))
        TemplateDomain.all_classes_selected = True

        cls.get_screen().set_all_classes_selected()

        cls.get_screen().update_all_classes_selected(cls.all_classes_selected)
        cls.get_screen().update_label_under_list(
            total_amount_of_items=cls.get_total_amount_of_classes(),
            counter=total_class_count)

    @classmethod
    def deselect_all_classes(cls):
        cls.selected_classes_indexes.clear()
        TemplateDomain.all_classes_selected = False

        cls.get_screen().deselect_all_classes()

        cls.get_screen().update_all_classes_selected(cls.all_classes_selected)
        cls.get_screen().update_label_under_list(total_amount_of_items=cls.get_total_amount_of_classes(),
                                                 counter=0)

    @classmethod
    def update_frontend(cls):
        cls.get_screen().update_project_info(global_vars.current_project)
        cls.get_screen().set_classes(classes=cls.classes,
                                     selected_classes= cls.selected_classes_indexes,
                                     all_classes_selected_checked=cls.all_classes_selected,
                                     has_a_class_with_deprecated_attributes=cls.has_a_class_with_deprecated_attributes)

    @classmethod
    def get_screen(cls) -> TemplateScreenInterface:
        return global_vars.otl_wizard.main_window.step1

    @classmethod
    @add_loading_screen
    async def async_export_template(cls, document_path:Path,
                        generate_choice_list: bool, geometry_column_added: bool,
                        export_attribute_info: bool, highlight_deprecated_attributes:bool,
                        amount_of_examples: int):
        return await cls.export_template(document_path,generate_choice_list,
                                   geometry_column_added,export_attribute_info,
                                   highlight_deprecated_attributes,amount_of_examples)

    @classmethod
    async def export_template(cls, document_path:Path ,
                        generate_choice_list: bool, geometry_column_added: bool,
                        export_attribute_info: bool, highlight_deprecated_attributes:bool,
                        amount_of_examples: int):

        project = global_vars.current_project

        selected_classes = [cls.classes[i].objectUri for i in cls.selected_classes_indexes]

        if not document_path or not project:
            return
        await TemplateDomain.create_template(subset_path=project.subset_path,
                                             document_path=Path(document_path),
                                             selected_classes_typeURI_list= selected_classes,
                                             generate_choice_list=generate_choice_list,
                                             add_geo_artefact=geometry_column_added,
                                             add_attribute_info=export_attribute_info,
                                             highlight_deprecated_attributes=highlight_deprecated_attributes,
                                             amount_of_examples=amount_of_examples,
                                             )
        if ".xlsx" in str(document_path):
            if platform.system() == 'Linux':
                os.open(document_path, os.O_WRONLY)
            elif platform.system() == 'Windows':
                os.startfile(document_path)
            else:
                OTLLogger.logger.error("Opening a file on this OS is not implemented yet")
        else:
            cls.get_screen().open_folder_of_created_template(document_path)

    @classmethod
    def toggle_class_index(cls,index:int):
        if index in cls.selected_classes_indexes:
            cls.selected_classes_indexes.remove(index)
        else:
            cls.selected_classes_indexes.append(index)

        cls.get_screen().update_label_under_list(total_amount_of_items=cls.get_total_amount_of_classes(),
                                                 counter=len(cls.selected_classes_indexes))
        cls.update_all_classes_selected_state()

    @classmethod
    def get_total_amount_of_classes(cls):

        count = len(cls.classes)
        if cls.has_agent:
            count -= 1
        return count

    @classmethod
    def update_all_classes_selected_state(cls):
        cls.all_classes_selected = len(cls.selected_classes_indexes) == cls.get_total_amount_of_classes()
        cls.get_screen().update_all_classes_selected(cls.all_classes_selected)

    @classmethod
    def set_select_all_classes(cls, new_state:bool):

        if new_state:
            cls.select_all_classes()
            return

        total_amount_of_items = cls.get_total_amount_of_classes()
        counter = len(cls.selected_classes_indexes)
        if total_amount_of_items == counter:
            cls.deselect_all_classes()
            return

        cls.all_classes_selected = False

    @classmethod
    def set_selected_indexes(cls,new_indexes: list[int]):

        cls.selected_classes_indexes = new_indexes
        cls.get_screen().update_label_under_list(
            total_amount_of_items=cls.get_total_amount_of_classes(),
            counter=len(cls.selected_classes_indexes))
        cls.update_all_classes_selected_state()


from __future__ import annotations

import datetime
import json
import os
import shutil
import zipfile
from collections import defaultdict
from pathlib import Path
from typing import Optional, Union, cast

from otlmow_model.OtlmowModel.BaseClasses.RelationInteractor import RelationInteractor
from otlmow_model.OtlmowModel.Classes.ImplementatieElement.AIMObject import AIMObject
from otlmow_model.OtlmowModel.Classes.ImplementatieElement.RelatieObject import RelatieObject
from otlmow_visuals.PyVisWrapper import PyVisWrapper

from otlmow_gui.Domain import global_vars
from otlmow_gui.Domain.util.Helpers import Helpers
from otlmow_gui.Domain.database.ModelBuilder import ModelBuilder
from otlmow_gui.Domain.logger.OTLLogger import OTLLogger
from otlmow_gui.Domain.project.ProjectFile import ProjectFile
from otlmow_gui.Domain.enums import FileState
from otlmow_gui.Domain.ProgramFileStructure import ProgramFileStructure
from otlmow_gui.Domain.util.VisualisationStateTracker import VisualisationStateTracker
from otlmow_gui.Exceptions.ExcelFileUnavailableError import ExcelFileUnavailableError
from otlmow_gui.GUI.dialog_windows.LoadingImageWindow import add_loading_screen
from otlmow_gui.GUI.dialog_windows.NotificationWindow import NotificationWindow
from otlmow_gui.GUI.dialog_windows.ProjectExistsError import ProjectExistsError
from otlmow_gui.GUI.dialog_windows.YesOrNoNotificationWindow import YesOrNoNotificationWindow
from otlmow_gui.GUI.translation.GlobalTranslate import GlobalTranslate
from otlmow_gui.exception_handler.ExceptionHandlers import create_task_reraise_exception


class Project:
    """
    Normal Project Structure:
    │   <self.eigen_referentie>                 : folder
    │   │   project_details.json                : json-file
    │   │   saved_documents.json                : json-file
    │   │   <self.subset_path_name>.db          : sqlite file
    │   ├───project-files                       : folder
    │   │       <saved_project_files[0]>.xlsx   : xlsx,json,geojson,csv file
    │   └───quick_saves                         : folder
    │           quick_save-<date>_<time>.json   : json-file
    """

    project_details_filename = 'project_details.json'
    saved_documents_filename = "saved_documents.json"
    project_files_foldername = 'project_files'
    old_project_files_foldername = 'OTL-template-files' # for legacy compatibility
    quick_saves_foldername =  "quick_saves"
    visualisation_foldername = "visuals"
    visualisation_filename = "graph_visualisation.html"
    visualisation_python_support_data_filename = "visuals_python_support_data.json"

    max_days_quicksave_stored = 7
    quicksave_date_format = "%y%m%d_%H%M%S"

    def __init__(self, eigen_referentie: str, project_path: Path = None, subset_path: Path = None,
                 saved_documents_overview_path: Path = None, bestek: str = None,
                 laatst_bewerkt: datetime.datetime = None, last_quick_save: Optional[Path] = None,
                 subset_operator: Optional[Union[str, Path]] = None,
                 otl_version: Optional[Union[str, Path]] = None):

        self.project_path: Path = project_path
        if not self.project_path:
            self.project_path = Path(
                ProgramFileStructure.get_otl_wizard_projects_dir() / eigen_referentie)

        self.subset_path: Path = subset_path
        
        if isinstance(subset_operator,Path):
            subset_operator = subset_operator.stem
        self.subset_operator: Optional[str] = subset_operator
        
        if isinstance(otl_version,Path):
            otl_version = otl_version.stem
        self.otl_version: Optional[str] = otl_version

        if saved_documents_overview_path:
            if ".json" in str(saved_documents_overview_path.absolute()):
                self.saved_documents_overview_path: Path = saved_documents_overview_path
            else:
                self.saved_documents_overview_path: Path = saved_documents_overview_path / self.saved_documents_filename
        else:
            self.saved_documents_overview_path = self.project_path / self.saved_documents_filename

        self.quick_save_dir_path = Path(self.project_path / self.quick_saves_foldername)

        if last_quick_save:
            self.last_quick_save:Path = self.get_quicksaves_dir_path() / last_quick_save.name
        else:
            self.last_quick_save = None

        self.eigen_referentie: str = eigen_referentie
        self.bestek: str = bestek
        self.laatst_bewerkt: datetime.datetime = laatst_bewerkt

        self.assets_in_memory:list[Union[RelatieObject, RelationInteractor]] = []

        self.saved_project_files: list[ProjectFile] = []
        self.model_builder = None
        #status tracking if the visualisation is uptodate with changes in relations and OTL-assets
        self.visualisation_uptodate: VisualisationStateTracker = VisualisationStateTracker()
        self.graph_saved_status: bool = True # status tracking if graph changes are written to html

    @classmethod
    def load_project(cls, project_path: Path = None):
        """
        Loads a project from a specified directory by reading its details from a JSON file.

        It validates the existence of the project directory and the required project details file
        before extracting and returning the project information.

        :param project_path: The path to the project directory. Defaults to None.
        :type project_path: Path, optional

        :return: An instance of the project class initialized with the loaded details.
        :rtype: Project

        :raises FileNotFoundError:  If the project directory or the project details file does not
                                    exist.
        """
        OTLLogger.logger.info(f"loading project details: {project_path.name}",extra={"timing_ref":f"details_{project_path.name}"})
        if project_path and not project_path.exists():
            raise FileNotFoundError(f"Project dir {project_path} does not exist")

        project_details_file = project_path / cls.project_details_filename
        if not project_details_file.exists():
            raise FileNotFoundError(f"Project details file {project_details_file} does not exist")

        with open(project_details_file) as json_file:
            project_details = json.load(json_file)

        if 'last_quick_save' not in project_details or not project_details['last_quick_save']:
            last_quick_save = None
        else:
            last_quick_save = (Path(project_path / cls.quick_saves_foldername) /
                               Path(project_details['last_quick_save']))


        if 'subset_operator' not in project_details or not project_details['subset_operator']:
            subset_operator = None
        else:
            subset_operator = project_details['subset_operator']

        if 'otl_version' not in project_details or not project_details['otl_version']:
            otl_version = None
        else:
            otl_version = project_details['otl_version']

        OTLLogger.logger.info(f"loading project details: {project_path.name}"
                              ,extra={"timing_ref":f"details_{project_path.name}"})
        return cls(project_path=project_path,
                   subset_path=project_path / project_details['subset'],
                   saved_documents_overview_path= Path(project_path,  cls.saved_documents_filename),
                   eigen_referentie=project_details['eigen_referentie'],
                   bestek=project_details['bestek'],
                   laatst_bewerkt=datetime.datetime.strptime(project_details['laatst_bewerkt'], "%Y-%m-%d %H:%M:%S"),
                   last_quick_save=last_quick_save,
                   subset_operator=subset_operator,
                   otl_version=otl_version)


    def save_project_to_dir(self) -> None:
        """
        Saves the current project to a designated directory, including project details and any
        associated subset files.

        This function ensures that the project directory is created if it does not exist and
        updates the last modified timestamp if necessary.

        :param self: The instance of the class containing project details.

        :return: None

        :raises OSError:    If there is an error creating the directory or writing the project
                            details file.

        :example:
            project.save_project_to_dir()
        """

        project_dir_path = self.get_project_local_path()

        OTLLogger.logger.debug("Saving project to %s", project_dir_path)
        project_dir_path.mkdir(exist_ok=True, parents=True)
        self.save_project_details(project_dir_path=project_dir_path)

        if self.subset_path and self.subset_path.parent.absolute() != project_dir_path.absolute():
            # move subset to project dir
            new_subset_path = project_dir_path / self.subset_path.name
            shutil.copy(src=self.subset_path, dst=new_subset_path)

    def get_project_local_path(self) -> Path:
        otl_wizard_project_dir = ProgramFileStructure.get_otl_wizard_projects_dir()
        return otl_wizard_project_dir / self.project_path.name

    def get_current_visuals_folder_path(self) -> Path:
        visuals_folder = self.get_project_local_path() / self.visualisation_foldername

        if not visuals_folder.exists():
            os.makedirs(visuals_folder, exist_ok=True)

        return visuals_folder

    def get_current_visuals_html_path(self) -> Path:
        return self.get_current_visuals_folder_path() / self.visualisation_filename

    def get_visualisation_python_support_data_path(self) -> Path:
        visuals_uptodate_state_path = self.get_current_visuals_folder_path() / self.visualisation_python_support_data_filename

        if not visuals_uptodate_state_path.exists():
            with open(visuals_uptodate_state_path,"w") as file:
                json.dump({"visuals_uptodate":False},file)

        return visuals_uptodate_state_path

    def load_visualisation_python_support_data(self,vis_wrap: Optional[PyVisWrapper] = None) -> None:
        file_path = self.get_visualisation_python_support_data_path()

        with open(file_path, "r") as file:
            json_data = json.load(file)

            self.visualisation_uptodate.set_clear_all(not json_data["visuals_uptodate"])
            if vis_wrap and "collection_support_data" in json_data.keys():
                vis_wrap.special_edges = json_data["collection_support_data"]["vis_wrap.special_edges"]
                vis_wrap.asset_id_to_display_name_dict = json_data["collection_support_data"][
                    "vis_wrap.asset_id_to_display_name_dict"]
                vis_wrap.relation_id_to_collection_id = defaultdict(list,json_data["collection_support_data"][
                    "vis_wrap.relation_id_to_collection_id"])
                vis_wrap.collection_id_to_list_of_relation_ids = defaultdict(list,json_data["collection_support_data"][
                    "vis_wrap.collection_id_to_list_of_relation_ids"])
                vis_wrap.collection_relation_count_threshold = json_data["collection_support_data"][
                    "vis_wrap.collection_relation_count_threshold"]

    def save_visualisation_python_support_data(self,vis_wrap: Optional[PyVisWrapper] = None) -> None:
        file_path = self.get_visualisation_python_support_data_path()

        if file_path.exists():
            # load file first to preserve previously saved data
            with open(file_path, "r") as file:
                json_data = json.load(file)
        else:
            json_data = {}


        json_data["visuals_uptodate"] = not self.visualisation_uptodate.get_clear_all()
        if vis_wrap:
            if "collection_support_data" not in json_data.keys():
                json_data["collection_support_data"] = {}

            json_data["collection_support_data"]["vis_wrap.special_edges"] = vis_wrap.special_edges
            json_data["collection_support_data"]["vis_wrap.asset_id_to_display_name_dict"] = vis_wrap.asset_id_to_display_name_dict
            json_data["collection_support_data"]["vis_wrap.relation_id_to_collection_id"] = vis_wrap.relation_id_to_collection_id
            json_data["collection_support_data"]["vis_wrap.collection_id_to_list_of_relation_ids"] = vis_wrap.collection_id_to_list_of_relation_ids
            json_data["collection_support_data"]["vis_wrap.collection_relation_count_threshold"] = vis_wrap.collection_relation_count_threshold
            # json_data["visualisation_option"] =

        with open(file_path, "w") as file:
            json.dump(json_data ,file)

    def save_project_details(self, project_dir_path: Path) -> None:
        """
        Saves the project's details to a specified directory in JSON format.

        This function constructs a dictionary of project attributes and writes it to a file, updating the last modified timestamp if necessary.

        :param self: The instance of the class managing the project.
        :param project_dir_path: The directory path where the project details will be saved.
        :type project_dir_path: Path

        :return: None

        :raises OSError: If there is an error writing the project details file.

        :example:
            project.save_project_details(Path('/path/to/project'))
        """

        last_quick_save_name = None
        if self.last_quick_save:
            last_quick_save_name = str(self.last_quick_save.name)

        if not self.laatst_bewerkt:
            self.laatst_bewerkt = datetime.datetime.now()

        subset_name = None
        if self.subset_path:
            subset_name = self.subset_path.name

        project_details_dict = {
            'bestek': self.bestek,
            'eigen_referentie': self.eigen_referentie,
            'laatst_bewerkt': self.laatst_bewerkt.strftime("%Y-%m-%d %H:%M:%S"),
            'subset': subset_name,
            'subset_operator': self.get_operator_name(),
            'otl_version': self.get_otl_version(),
            'last_quick_save': last_quick_save_name
        }

        with open(project_dir_path / self.project_details_filename, "w") as project_details_file:
            json.dump(project_details_dict, project_details_file)

    @add_loading_screen
    async def load_validated_assets_async(self) -> list[Union[RelatieObject, RelationInteractor]]:
        # sourcery skip: assign-if-exp, reintroduce-else, use-named-expression
        """
        Loads validated assets from the most recent quick save file.

        If a valid quick save path is found, it converts the file contents into a list of AIMObject
        instances; otherwise, it returns an empty list.

        :param self: The instance of the class managing asset loading.

        :return:    A list of validated AIMObject instances loaded from the quick save file, or an
                    empty list if no valid path is found.
        :rtype: list[AIMObject]

        :example:
            assets = project.load_validated_assets()
        """

        path = self.get_last_quick_save_path()

        if path:
            timing_ref = f"load_assets_{path.stem}"
            OTLLogger.logger.debug(
                f"Execute Project.load_validated_assets({path.name}) for project "
                f"{self.eigen_referentie}",
                extra={"timing_ref": timing_ref})

            # noinspection PyTypeChecker
            saved_objects, exceptions_group = await Helpers.converter_from_file_to_object_async(
                path,
                allow_non_otl_conform_attributes=True)

            object_count = len(saved_objects)
            OTLLogger.logger.debug(
                f"Execute Project.load_validated_assets({path.name}) for project"
                f" {self.eigen_referentie} ({object_count} objects)",
                extra={"timing_ref":  timing_ref})

            return saved_objects
        return []


    def load_validated_assets(self) -> list[Union[RelatieObject, RelationInteractor]]:
        # sourcery skip: assign-if-exp, reintroduce-else, use-named-expression
        """
        Loads validated assets from the most recent quick save file.

        If a valid quick save path is found, it converts the file contents into a list of AIMObject
        instances; otherwise, it returns an empty list.

        :param self: The instance of the class managing asset loading.

        :return:    A list of validated AIMObject instances loaded from the quick save file, or an
                    empty list if no valid path is found.
        :rtype: list[AIMObject]

        :example:
            assets = project.load_validated_assets()
        """

        path = self.get_last_quick_save_path()

        if path:
            timing_ref = f"load_assets_{path.stem}"
            OTLLogger.logger.debug(
                f"Execute Project.load_validated_assets({path.name}) for project "
                f"{self.eigen_referentie}",
                extra={"timing_ref": timing_ref})

            # noinspection PyTypeChecker
            saved_objects, exceptions_group = Helpers.converter_from_file_to_object(
                path)

            object_count = len(saved_objects)
            OTLLogger.logger.debug(
                f"Execute Project.load_validated_assets({path.name}) for project"
                f" {self.eigen_referentie} ({object_count} objects)",
                extra={"timing_ref": timing_ref})

            return saved_objects
        return []

    async def save_validated_assets(self, asynchronous=True) -> None:
        """
        Saves validated assets to the project quick save directory.

        It manages the quick save files by removing old saves and generating a new save file with a unique name.

        The format of the save file is: quick_save_{date}_{time}.json:
            date:   the date on which the quicksave was made in format yymmdd
            time:   the time on which the quicksave was made in format HHMMSS

        All quicksaves that are more than set days old (default= 7) are deleted in this function.

        :param self: The instance of the class managing quick saves.

        :return: None

        :raises OSError: If there is an issue creating the quick save directory or saving the file.
        @param asynchronous: do the save asynchronous or not
        """

        if self.quick_save_dir_path.exists():
            current_date = datetime.datetime.now()

            self.remove_too_old_quicksaves(current_date=current_date,
                                          max_days_stored=Project.max_days_quicksave_stored,
                                          date_format=Project.quicksave_date_format)
        else:
            os.mkdir(self.quick_save_dir_path )

        current_date_str = datetime.datetime.now().strftime(Project.quicksave_date_format)

        save_path = self.quick_save_dir_path  / f"quick_save-{current_date_str}.json"
        if asynchronous:
            try:
                create_task_reraise_exception(self.make_quick_save_async(save_path=save_path))
            except DeprecationWarning:
                # should only go here if you are testing
                self.make_quick_save(save_path=save_path)
        else:
            self.make_quick_save(save_path=save_path)


    async def make_quick_save_async(self, save_path: Path) -> None:
        object_count = len(self.assets_in_memory)
        OTLLogger.logger.debug(f"Execute Project.make_quick_save({save_path.name}) for project {self.eigen_referentie} ({object_count} objects)", extra={
            "timing_ref": f"make_quick_save_{save_path.stem}"})

        await Helpers.from_objects_to_file_async(save_path,self.assets_in_memory)
        self.last_quick_save = save_path
        self.save_project_to_dir()
        OTLLogger.logger.debug(
            f"Execute Project.make_quick_save({save_path.name}) for project {self.eigen_referentie} ({object_count} objects)",
            extra={
                "timing_ref": f"make_quick_save_{save_path.stem}"})



    def make_quick_save(self, save_path: Path) -> None:
        object_count = len(self.assets_in_memory)
        OTLLogger.logger.debug(f"Execute Project.make_quick_save({save_path.name}) for project {self.eigen_referentie} ({object_count} objects)", extra={
            "timing_ref": f"make_quick_save_{save_path.stem}"})

        Helpers.from_objects_to_file(save_path,self.assets_in_memory)
        self.last_quick_save = save_path
        self.save_project_to_dir()
        OTLLogger.logger.debug(
            f"Execute Project.make_quick_save({save_path.name}) for project {self.eigen_referentie} ({object_count} objects)",
            extra={
                "timing_ref": f"make_quick_save_{save_path.stem}"})



    def remove_too_old_quicksaves(self, current_date: datetime, max_days_stored: datetime,
                                   date_format: str) -> None:
        """
        Removes quick save files that are older than a specified number of days.

        This function checks the modification dates of the quick save files and deletes those that exceed the maximum age defined by the user.

        Files that do not adhere to the standard naming convention are never deleted.

        :param self: The instance of the class managing quick saves.
        :param current_date: The current date used to compare against quick save file dates.
        :type current_date: datetime
        :param max_days_stored: The maximum number of days to retain quick save files.
        :type max_days_stored: datetime
        :param date_format: The format used to parse the date from the filename.
        :type date_format: str

        :return: None

        :example:
            project.remove_too_old_quicksaves(datetime.now(), 30, "%Y-%m-%d")
        """


        files = os.listdir(path=self.quick_save_dir_path )
        for filename in files:
            try:
                file_date = datetime.datetime.strptime(filename.split("-")[-1].split(".json")[0],
                                                       date_format)
            except ValueError:
                # if the save file doesn't adhere to standard naming convention it is never deleted
                file_date = current_date

            days_ago = (current_date - file_date).days
            if days_ago > max_days_stored:
                file_to_remove_path = Path(self.quick_save_dir_path , filename)
                os.remove(file_to_remove_path)


    def export_project_to_file(self, file_path: Path) -> None:
        """
        Exports the project data to a specified file in OTLW format which is a custom ZIP format.

        This method creates a ZIP file containing the project's details,
        including the project details file, saved documents overview, subset path, and any saved
        project files. It also includes the last quick save file if it exists, ensuring that all
        relevant project information is packaged together.

        :param file_path: The path where the OTLW file will be saved.
        :type file_path: Path

        :return: None

        :raises FileNotFoundError: If any of the files to be included in the ZIP do not exist.
        """

        with zipfile.ZipFile(file=file_path, mode='w') as project_zip:
            project_zip.write(self.project_path / self.project_details_filename,
                              arcname=self.project_details_filename,
                              compresslevel=zipfile.ZIP_DEFLATED)
            if self.saved_documents_overview_path and self.saved_documents_overview_path.exists():
                project_zip.write(self.saved_documents_overview_path,
                                  arcname=self.saved_documents_overview_path.name)
            project_zip.write(self.subset_path, arcname=self.subset_path.name)

            if not self.get_saved_projectfiles():
                self.load_saved_document_filenames()

            # tell the user there are files missing
            missing_project_files = self.check_if_project_files_exist()
            if len(missing_project_files):
                message = GlobalTranslate._(
                    "There are files missing that belong to this project:\n")
                for project_file in missing_project_files:
                    filename = project_file.file_path.name
                    message += f"   -{filename}\n"

                message += GlobalTranslate._("Do you still want to export?")
                msgbox = YesOrNoNotificationWindow(message=message, title=GlobalTranslate._("Missing project files"))
                answer = msgbox.exec()

                if answer == 65536:
                    return

            for document in self.get_saved_projectfiles():
                if document.file_path.exists():
                    file_zip_path = Path(self.project_files_foldername) / document.file_path.name
                    project_zip.write(document.file_path, arcname=file_zip_path)

            last_quick_save_path = self.get_last_quick_save_path()
            if last_quick_save_path and last_quick_save_path.exists():
                last_quick_save_zip_path = Path(self.quick_saves_foldername) / last_quick_save_path.name
                project_zip.write(last_quick_save_path, arcname=last_quick_save_zip_path)

            visualisation_html_path = self.get_current_visuals_html_path()
            if visualisation_html_path and visualisation_html_path.exists():
                visualisation_html_zip_path = Path(
                    self.visualisation_foldername) / visualisation_html_path.name
                project_zip.write(visualisation_html_path, arcname=visualisation_html_zip_path)

            visualisation_uptodate_json_path = self.get_visualisation_python_support_data_path()
            if visualisation_uptodate_json_path and visualisation_uptodate_json_path.exists():
                visualisation_uptodate_json_zip_path = Path(
                    self.visualisation_foldername) / visualisation_uptodate_json_path.name
                project_zip.write(visualisation_uptodate_json_path, arcname=visualisation_uptodate_json_zip_path)

    @classmethod
    def import_project(cls, file_path: Path):
        """
        Imports a project from a specified OTLW file which is a custom ZIP format.

        This class method extracts project details from a ZIP file and creates a new project directory based on the extracted information. It handles the creation of the project directory and raises an error if the directory already exists.

        :param file_path: The path to the OTLW file containing the project data.
        :type file_path: Path

        :return: An instance of the Project class initialized with the imported project details.
        :rtype: Self

        :raises FileExistsError: If the project directory already exists.
        """

        with zipfile.ZipFile(file=file_path) as project_file:

            # TODO: handle if project doesn't contain project_details.json files
            project_details = json.load(fp=project_file.open(cls.project_details_filename))

            project_dir_path = Path(
                ProgramFileStructure.get_otl_wizard_projects_dir() / project_details[
                    'eigen_referentie'])
            try:
                project_dir_path.mkdir(exist_ok=False, parents=True)
            except FileExistsError as ex:
                OTLLogger.logger.error("Project dir %s already exists", project_dir_path)
                raise ProjectExistsError(eigen_referentie=project_details['eigen_referentie'])

            project_file.extractall(path=project_dir_path)

        return Project.load_project(project_path=project_dir_path)

    def delete_project_dir_by_path(self) -> None:
        """
        Deletes the project directory specified by the project's path.

        This method removes the entire directory associated with the project, including all its contents. It logs the deletion action for debugging purposes.

        :return: None
        """

        OTLLogger.logger.debug("Deleting project %s", self.project_path)
        shutil.rmtree(path=self.project_path)



    def are_all_project_files_in_memory_valid(self) -> bool:
        """
        Checks if all project files currently in memory are valid.

        This method evaluates the state of saved project files to determine if at least one file
        meets the required validity criteria. It logs the process and returns a boolean indicating
        whether any project files are valid.

        :return: True if at least one project file is valid; otherwise, False.
        :rtype: bool
        """

        # OTLLogger.logger.debug("Started searching for project files in memory that are OTL conform")

        if not self.saved_project_files:
            OTLLogger.logger.debug("No project files in memory")
            return False
        return all(
            template.state == FileState.OK for template in self.saved_project_files
        )

    def get_subset_db_name(self) -> str:
        """
        Retrieves the name of the subset database associated with the project.

        This function returns the name of the subset path as a string.

        :return: The name of the subset database.
        :rtype: str
        """

        return  str(self.subset_path.name)

    def get_model_builder(self) -> ModelBuilder:
        """
        Loads and initializes the model builder associated with the project if it has not been
        created yet.

        This function checks if the model builder is already set and creates a new instance using
        the subset path if necessary.

        :return: The model builder instance associated with the project.
        :rtype: ModelBuilder

        :raises ValueError: If the subset path is not set when attempting to create the model
                            builder.
        """


        if not self.model_builder and self.subset_path:
            self.model_builder = ModelBuilder(self.subset_path)
        return self.model_builder

    def clear_model_builder_from_memory(self) -> None:
        """
        Clears the model builder from memory by setting it to None.

        This function effectively releases the reference to the current model builder, allowing for garbage collection.

        :return: None
        """

        self.model_builder = None

    def get_operator_name(self) -> str:
        """
        Retrieves the name of the subset operator associated with the project.

        If the subset operator is not already set, it fetches the operator name from the model builder and processes it accordingly.

        :return: The name of the subset operator.
        :rtype: str

        :raises ValueError: If the operator name cannot be determined from the model builder.
        """

        if self.subset_path and not self.subset_operator:
            subset_operator = self.get_model_builder().get_operator_name()
            if isinstance(subset_operator,Path):
                self.subset_operator = cast(Path,subset_operator).stem
            else:
                self.subset_operator = subset_operator
        return self.subset_operator

    def get_otl_version(self) -> str:
        """
        Retrieves the OTL (Object Type Library) version associated with the project.

        If the OTL version is not already set, it fetches the version from the model builder.

        :return: The OTL version.
        :rtype: str

        :raises ValueError: If the OTL version cannot be determined from the model builder.
        """

        if self.subset_path and not self.otl_version:
            self.otl_version = self.get_model_builder().get_otl_version()


        return self.otl_version

    def load_saved_document_filenames(self):
        """
        Loads the filenames of saved documents associated with the project from a specified directory.

        This function checks for the existence of a saved documents file, reads its contents, and populates the list of saved project files with their saved states. If there are no quicksaves yet then the states are set to FileState.WARNING.

        :return: The instance of the class, allowing for method chaining.
        :rtype: self

        :raises FileNotFoundError: If the saved documents file does not exist.
        :raises json.JSONDecodeError: If the saved documents file contains invalid JSON.
        """

        project_dir_path = ProgramFileStructure.get_otl_wizard_projects_dir() / self.project_path.name
        saved_documents_path: Path = project_dir_path / Project.saved_documents_filename



        if saved_documents_path.exists():

            OTLLogger.logger.debug(
                f"Loading saved documents: {self.project_path.name}/{Project.saved_documents_filename}",
                extra={"timing_ref": f"load_quicksave_{self.project_path.name}"})

            with open(file=saved_documents_path,mode="r") as saved_document:
                try:
                    saved_documents = json.load(fp=saved_document)
                except json.decoder.JSONDecodeError as e:
                    # happens when the saved_documents.json file is empty
                    OTLLogger.logger.warning(e)
                    saved_documents = []
            saved_documents_str = str(saved_documents)

            self.saved_project_files = []

            location_dir = self.get_project_files_dir_path()

            #copy all files from the old OTL-Template-files folder to the new folder
            if not location_dir.exists():
                location_dir.mkdir()

            old_path = self.get_old_project_files_dir_path()
            if old_path.exists():
                for location, dirs, files in os.walk(old_path):
                    for file in files:
                        old_location = Path(location, file)
                        new_location = Path(location_dir , file)
                        shutil.copyfile(old_location, new_location)
                        os.remove(old_location)
                old_path.rmdir()

            for document in saved_documents:
                # if document was validated but there is no quicksave then set it to warning
                state = FileState(document['state'])
                quick_save_path = self.get_quicksaves_dir_path()
                document_file_path = location_dir / Path(document['file_path']).name

                if (state == FileState.OK and
                    not (quick_save_path.exists() and
                     len(list(os.listdir(path=quick_save_path))))):
                    state = FileState.WARNING

                file = ProjectFile(
                    file_path=location_dir / Path(document['file_path']).name ,
                    state=state)
                self.saved_project_files.append(file)

            OTLLogger.logger.debug(
                f"Loading saved documents: {self.project_path.name}/{Project.saved_documents_filename}",
                extra={"timing_ref": f"load_quicksave_{self.project_path.name}"})
        return self

    def check_if_project_files_exist(self) -> list[ProjectFile]:
        missing_project_files = []
        for project_file in self.get_saved_projectfiles():
            if not project_file.file_path.exists():
                OTLLogger.logger.info(f'project "{self.eigen_referentie}" is missing file: {project_file.file_path}')
                missing_project_files.append(project_file)
                project_file.state = FileState.ERROR


        return missing_project_files

    def get_quicksaves_dir_path(self) -> Path:
        """
        Retrieves the path to the quick saves directory for the project.

        If the directory does not exist, it creates it before returning the path.

        :return: The path to the quick saves directory.
        :rtype: Path

        :raises OSError: If the directory cannot be created due to a file system error.
        """

        quick_saves = Path(self.project_path / self.quick_saves_foldername)
        if not quick_saves.exists() and self.project_path.exists():
            os.mkdir(Path(self.project_path / self.quick_saves_foldername))
        return quick_saves

    def get_last_quick_save_path(self) -> Optional[Path]:
        # sourcery skip: use-named-expression
        """
        Retrieves the path to the last quick save file for the project.

        It first checks if there is a specific last quick save path and if that exists; if not, it looks for the most recent file in the quick saves directory.

        :return: The path to the last quick save file, or None if no quick save exists.
        :rtype: Optional[Path]

        :raises OSError: If there is an issue accessing the quick saves directory.
        """

        path = None
        quick_save_path_dir = self.get_quicksaves_dir_path()
        if self.last_quick_save and self.last_quick_save.exists():
            path = self.last_quick_save
        elif quick_save_path_dir.exists():
            file_list = sorted(os.listdir(quick_save_path_dir), reverse=True)
            if file_list:
                path = Path(quick_save_path_dir, file_list[0])
        return path

    def get_project_files_dir_path(self) -> Path:
        """
        Retrieves the directory path for the project's files.

        This function constructs and returns the full path to the folder designated for storing project files.

        :param self: The instance of the class managing the project.

        :return: The path to the project files directory.
        :rtype: Path

        :example:
            files_dir = project.get_project_files_dir_path()
        """

        return self.project_path / self.project_files_foldername

    def get_old_project_files_dir_path(self) -> Path:
        """
        Retrieves the directory path for the old project's files.

        This function constructs and returns the full path to the folder designated for storing project files in projects made with an old version of OTL wizard.

        :param self: The instance of the class managing the project.

        :return: The path to the old project files directory.
        :rtype: Path

        :example:
            old_files_dir = project.get_old_project_files_dir_path()
        """

        return self.project_path / self.old_project_files_foldername

    def delete_template_folder(self) -> None:
        """
        Deletes the template folders associated with the project.

        This method removes both the old and new project files directories, effectively clearing
        out all template files. It logs the start and completion of the deletion process for
        debugging purposes.

        :return: None
        """

        OTLLogger.logger.debug(f"Started clearing out the whole template folder of "
                      f"project: {self.eigen_referentie}")
        new_project_files_folder = self.get_project_files_dir_path()
        old_project_files_folder = self.get_old_project_files_dir_path()

        if old_project_files_folder.exists():
            shutil.rmtree(path=old_project_files_folder)

        if new_project_files_folder.exists():
            shutil.rmtree(path=new_project_files_folder)

        OTLLogger.logger.debug("Finished clearing out the whole template folder")

    def save_project_filepaths_to_file(self) -> None:
        """
        Saves the file paths and states of the project's saved files to a JSON file.

        This function constructs a list of saved project file details and writes them to a specified file within the project's directory. It will overwrite the existing saved_documents.json.

        :return: None

        :raises OSError: If there is an issue writing to the file or accessing the project directory.
        """

        otl_wizard_project_dir = ProgramFileStructure.get_otl_wizard_projects_dir()
        object_array = []
        for objects_list in self.saved_project_files:
            objects_list_details = {
                'file_path': str(objects_list.file_path.name),
                'state': objects_list.state.value
            }
            object_array.append(objects_list_details)
        project_dir_path = otl_wizard_project_dir / self.project_path.name
        with open(file=project_dir_path / self.saved_documents_filename, mode="w+") as project_details_file:
            json.dump(object_array, project_details_file)

    def is_in_project(self, file_path) -> ProjectFile:
        """
        Checks if a specified file path is present in the project's saved files.

        This function searches through the list of saved project files and returns the matching project file if found, or None if no match exists.

        :param file_path: The path of the file to check for in the project's saved files.
        :type file_path: Path

        :return: The matching project file if found, or None if not.
        :rtype: ProjectFile

        :raises TypeError: If the provided file_path is not of the expected type.
        """

        matches = [project_file for project_file in self.saved_project_files if
         project_file.file_path == file_path]
        return  matches[0] if matches else None

    def get_saved_projectfiles(self) -> list[ProjectFile]:
        """
        Retrieves the list of saved project files associated with the project.

        This function returns the current state of the saved project files as a list of ProjectFile instances.

        :return: A list containing the saved project files.
        :rtype: list[ProjectFile]
        """

        return self.saved_project_files

    def copy_and_add_project_file(self, file_path: Path | str, state: FileState) -> None:
        """
        Adds a saved project file to the list of saved project files, makes a copy of it, and
        updates its state.

        This function converts the file path to a Path object if it is provided as a string,
        constructs the full file path, and appends a new ProjectFile instance to the saved project
        files.

        :param file_path: The path of the file to be saved, which can be a string or a Path object.
        :type file_path: Path | str
        :param state: The state of the project file to be saved.
        :type state: FileState

        :return: None

        :raises ValueError: If the provided file_path is invalid or if the state is not a valid
                            FileState.
        """


        if isinstance(file_path,str):
            file_path = Path(file_path)

        if self.has_duplicate_filename(filename=file_path.name,
                                       project_files_list=self.saved_project_files):
            return

        end_location = self.make_copy_of_added_file(filepath=file_path)
        self.saved_project_files.append(ProjectFile(file_path=end_location, state=state))
        self.save_project_filepaths_to_file()

    def has_duplicate_filename(self, filename: str, project_files_list: list[ProjectFile]) -> bool:
        file_names_in_project = [project_file.file_path.name for project_file in
                                 project_files_list]
        has_duplicate = False
        if filename in file_names_in_project:
            message = GlobalTranslate._("Can't insert duplicate filename {0}")

            msgbox = NotificationWindow(message=message.format(filename),
                                        title=GlobalTranslate._("Duplicate filename"))
            msgbox.exec()
            has_duplicate = True
        return has_duplicate

    def make_copy_of_added_file(self, filepath: Path) -> Path:
        """
        Creates a copy of the specified file in the OTL template files directory.

        This function ensures the destination directory exists, copies the file if it is not
        already in the target location, and returns the path to the copied file.

        :param filepath: The path of the file to be copied.
        :type filepath: Path

        :return: The path to the copied file in the OTL template files directory.
        :rtype: Path

        :raises FileNotFoundError: If the specified file does not exist.
        :raises OSError: If there is an issue creating the directory or copying the file.
        """

        location_dir = self.get_project_files_dir_path()
        if not location_dir.exists():
            old_path = self.get_old_project_files_dir_path()
            if old_path.exists():
                location_dir = old_path
            else:
                location_dir.mkdir()

        doc_name = filepath.name
        end_location = location_dir / doc_name
        if end_location == filepath:
            return end_location
        shutil.copy(src=filepath, dst=end_location)
        OTLLogger.logger.debug(f"Created a copy of the template file {filepath.name} in the project folder")
        return end_location


    def remove_all_project_files(self):
        """
        Removes all project files from memory and deletes them from the file system.

        This function iterates through the list of saved project files, attempts to delete each one, and handles any errors that occur during the deletion process. It creates a notification window to the user if the program doesn't have access to the file, and an empty saved_project_files list is saved to saved_document.json after the whole process.

        :return: None

        :raises ExcelFileUnavailableError: If a file cannot be deleted due to its unavailability.
        """

        OTLLogger.logger.debug("memory contains %s", self.saved_project_files)
        for file in self.saved_project_files:
            OTLLogger.logger.debug("starting to delete file %s",file.file_path)
            try:
                self._delete_project_file(file_path=file.file_path)
            except ExcelFileUnavailableError as e:
                # TODO: make a call here to GUI to create this dialogwindow
                global_vars.otl_wizard.main_window.notify_user_of_excel_file_unavailable_error(e)

        self.saved_project_files = []
        self.save_project_filepaths_to_file()


    def remove_project_file(self, file_path) -> bool:
        """
        Removes a specified project file from the list of saved project files and deletes it from
        the file system if it exists.

        This function checks if the file is part of the project, attempts to delete it, and
        updates the saved project files accordingly.

        :param file_path: The path of the project file to be removed.
        :type file_path: Path | str

        :return: True if the file was successfully removed, False otherwise.

        :raises ExcelFileUnavailableError: If the file cannot be deleted due to its unavailability.
        """

        project_file = self.is_in_project(file_path=file_path)
        res = False
        if project_file:
            if project_file.file_path.exists():
                try:
                    res =  self._delete_project_file(file_path=project_file.file_path)
                except ExcelFileUnavailableError as e:
                    #TODO: make a call here to GUI to create this dialogwindow
                    global_vars.otl_wizard.main_window.notify_user_of_excel_file_unavailable_error(e)
            else:
                res = True

            if res:
                self.saved_project_files.remove(project_file)
                self.save_project_filepaths_to_file()
        return res

    @staticmethod
    def _delete_project_file(file_path) -> bool:
        """
        Deletes a specified project file from the file system.

        This function attempts to unlink the file at the given path and handles some errors that may occur during the deletion process. Needs following calls after calling this one: self.saved_project_files.remove(project_file) and self.save_project_filepaths_to_file().

        :param file_path: The path of the project file to be deleted.
        :type file_path: Path | str

        :return: True if the file was successfully deleted, False if the file was not found.

        :raises ExcelFileUnavailableError: If there is a permission error while attempting to delete the file.
        """

        try:
            file_path_str = str(file_path)
            OTLLogger.logger.debug(f"file path = {file_path_str}")
            Path(file_path).unlink()
            return True
        except FileNotFoundError as e:
            OTLLogger.logger.error(e)
            return False
        except PermissionError as e:
            #TODO: make other errors + dialog windows for cases with non-excel files if necessary
            OTLLogger.logger.error(e)
            raise ExcelFileUnavailableError(file_path=file_path, exception=e) from e

    def change_subset(self, new_path: Path) -> None:
        """
        Changes the current subset path for the project and resets related properties.

        This function clears the model builder from memory, updates the subset path, loads a new model builder, and resets the subset operator and OTL version to their initial states.

        :param new_path: The new path to set as the subset for the project.
        :type new_path: Path

        :return: None
        """

        self.clear_model_builder_from_memory()
        self.subset_path = new_path
        self.model_builder = self.get_model_builder()
        self.subset_operator = None
        self.otl_version = None
        self.update_last_alter_time()
        self.save_project_to_dir()

    def update_information(self, new_eigen_ref:str, new_bestek:str, new_subset_path: Path):
        """
        Updates the project information with new details.

        This method modifies the project's eigen reference,
        bestek, and subset path, ensuring that the project reflects the latest information.
        It also updates the last altered time to track when the changes were made.

        :param new_eigen_ref: The new eigen reference for the project.
        :type new_eigen_ref: str
        :param new_bestek: The new bestek for the project.
        :type new_bestek: str
        :param new_subset_path: The new path for the project's subset.
        :type new_subset_path: Path

        :return: None
        """
        # check if there is a difference avoid unnecessary renaming of folders
        if new_eigen_ref != self.eigen_referentie:
            # include the new_subset_path so it can be changed to the new folder name if it is the old subset
            new_subset_path = self.update_eigen_referentie(new_eigen_ref,new_subset_path)
        self.bestek = new_bestek
        self.change_subset(new_subset_path)
        self.update_last_alter_time()


    def update_eigen_referentie(self,new_eigen_ref:str, new_subset_path:Path=None ) -> Path:
        new_project_path = Path(ProgramFileStructure.get_otl_wizard_projects_dir() / new_eigen_ref)
        try:
            if new_project_path.exists():
                raise FileExistsError
            shutil.move(self.project_path, new_project_path)
        except FileExistsError as ex:
            OTLLogger.logger.error("Project dir %s already exists", new_project_path)
            raise ProjectExistsError(eigen_referentie=new_eigen_ref)
        except PermissionError as ex:
            OTLLogger.logger.error("No permission to rename ", self.project_path)
            raise ex

        # try:
        #     shutil.copytree(self.project_path, new_project_path)
        # except FileExistsError as ex:
        #     OTLLogger.logger.error("Project dir %s already exists", new_project_path)
        #     raise ProjectExistsError(eigen_referentie=new_eigen_ref)

        # delete the original path
        # self.delete_project_dir_by_path()

        # change the new subset path otherwise change the old one
        if not new_subset_path:
            new_subset_path = self.subset_path
        elif new_subset_path == self.subset_path:
            new_subset_path = new_project_path / self.subset_path.name


        # make sure the current data is correctly renamed before updating the projects attributes
        self.eigen_referentie = new_eigen_ref
        self.project_path = new_project_path

        return new_subset_path


    def update_last_alter_time(self):
        """
        Updates the last altered time of the project.

        This method sets the last altered timestamp to the current date and time, allowing for tracking of when the project was last modified.

        :return: None
        """

        time_of_alter = datetime.datetime.now().date()
        self.laatst_bewerkt = time_of_alter


    def __eq__(self, __value):
        if not __value:
            return False

        return (self.eigen_referentie == __value.eigen_referentie and
                Helpers.equal_paths(path1=self.subset_path,path2=__value.subset_path) and
                self.subset_operator == __value.subset_operator and
                self.otl_version == __value.otl_version and
                self.bestek == __value.bestek and
                self.laatst_bewerkt == __value.laatst_bewerkt and
                self.saved_project_files == __value.saved_project_files)

    def set_saved_graph_status(self,new_status:bool) -> None:
        self.graph_saved_status = new_status

    def get_saved_graph_status(self) -> bool:
        return self.graph_saved_status




import base64
import os
import subprocess
import sys
import tempfile
import traceback
import warnings
from pathlib import Path
from typing import Optional, Iterable

from otlmow_converter.Exceptions.ExceptionsGroup import ExceptionsGroup
from otlmow_converter.Exceptions.UnknownExcelError import UnknownExcelError
from otlmow_converter.OtlmowConverter import OtlmowConverter
from otlmow_model.OtlmowModel.BaseClasses.OTLObject import OTLObject, \
    dynamic_create_instance_from_ns_and_name, dynamic_create_instance_from_uri

from otlmow_model.OtlmowModel.Helpers import OTLObjectHelper
from otlmow_model.OtlmowModel.Helpers.GenericHelper import validate_guid, get_ns_and_name_from_uri, \
    get_titlecase_from_ns
from otlmow_model.OtlmowModel.Helpers.generated_lists import get_hardcoded_class_dict, get_hardcoded_relation_dict
from packaging.version import Version

from otlmow_gui.Domain.logger.OTLLogger import OTLLogger
from otlmow_gui.GUI.dialog_windows.LoadingImageWindow import add_loading_screen
from otlmow_gui.GUI.screens.RelationChange_elements.RelationChangeHelpers import RelationChangeHelpers



class Helpers:
    all_OTL_asset_types_dict = {}

    @classmethod
    def get_hardcoded_class_dict(cls) -> dict:
        return get_hardcoded_class_dict()

    @classmethod
    def create_external_typeURI_options(cls):
        cls.all_OTL_asset_types_dict = {}
        all_type_uris = cls.get_hardcoded_class_dict()
        buckets_dict = {'assets': {}, 'legacy':{}}

        for uri, info in all_type_uris.items():
            if info['abstract']:
                continue
            if uri in get_hardcoded_relation_dict():
                continue

            ns, name = get_ns_and_name_from_uri(uri)
            screen_name = info['name'] if ns == 'legacy' else info['label']
            if ns is not None:
                screen_name += f" ({get_titlecase_from_ns(ns)})"

            if ns == 'legacy':
                screen_name = screen_name.replace("(Legacy) (Legacy)", "(Legacy)")
                buckets_dict['legacy'][screen_name] = uri
            else:
                buckets_dict['assets'][screen_name] = uri

            cls.all_OTL_asset_types_dict[screen_name] = uri

        # combine and sort the dictionary by keys (screen name)
        cls.all_OTL_asset_types_dict = dict(sorted(buckets_dict['assets'].items()))
        cls.all_OTL_asset_types_dict.update(dict(sorted(buckets_dict['legacy'].items())))

    @classmethod
    def sort_nested_dict(cls, dictionary, by='keys'):
        """Recursively sorts a dictionary by keys or values."""
        if not isinstance(dictionary, dict):
            if isinstance(dictionary, list):
                return sorted(dictionary, key=lambda relation_object: relation_object.typeURI)
            return dictionary

        # Sort the current dictionary
        if by == 'keys':
            sorted_dict = {
                k: cls.sort_nested_dict(v, by=by)
                for k, v in sorted(dictionary.items())
            }

        elif by == 'values':
            sorted_dict = {
                k: cls.sort_nested_dict(v, by=by)
                for k, v in sorted(dictionary.items(), key=lambda item: item[1])
            }

        else:
            raise ValueError("Invalid sort parameter. Use 'keys' or 'values'.")

        return sorted_dict
    
    @classmethod
    def equal_paths(cls,path1: Optional[Path], path2: Optional[Path]):
        if path1 and path2:
            return path1.name == path2.name
        elif path1 or path2:
            return False
        else:
            return True

    @classmethod
    async def converter_from_file_to_object_async(cls, file_path:Path,allow_non_otl_conform_attributes:bool=False, **kwargs):

        OTLLogger.logger.debug(f"Execute OtlmowConverter.from_file_to_objects({file_path.name})",
                               extra={"timing_ref": f"file_to_objects_{file_path.stem}"})

        exception_group = ExceptionsGroup(
            message=f'Failed to create objects')

        try:
            object_lists = list(await OtlmowConverter.from_file_to_objects_async(file_path,allow_non_otl_conform_attributes=allow_non_otl_conform_attributes,**kwargs))
        except ExceptionsGroup as group:
            exception_group = group
            object_lists = group.objects
        except BaseException as ex:
            object_lists = []
            exc_type, exc_value, exc_tb = sys.exc_info()
            tb = "".join(traceback.format_exception(exc_type, exc_value, exc_tb))
            OTLLogger.logger.error("error caught!")
            OTLLogger.logger.error("error message: \n: " + tb)
            exception_group.add_exception(UnknownExcelError(original_exception=ex, tab=""))

        object_count = len(object_lists)
        OTLLogger.logger.debug(f"Execute OtlmowConverter.from_file_to_objects({file_path.name}) ({object_count} objects)",
                               extra={"timing_ref": f"file_to_objects_{file_path.stem}"})
        return object_lists,exception_group

    @classmethod
    async def converter_from_file_to_object(cls, file_path, **kwargs):

        OTLLogger.logger.debug(f"Execute OtlmowConverter.from_file_to_objects({file_path.name})",
                               extra={"timing_ref": f"file_to_objects_{file_path.stem}"})
        exception_group = None
        try:
            object_lists = list(
                OtlmowConverter.from_file_to_objects(file_path, **kwargs))
        except ExceptionsGroup as group:
            exception_group = group
            object_lists = group.objects

        object_count = len(object_lists)
        OTLLogger.logger.debug(
            f"Execute OtlmowConverter.from_file_to_objects({file_path.name}) ({object_count} objects)",
            extra={"timing_ref": f"file_to_objects_{file_path.stem}"})
        return object_lists, exception_group

    @classmethod
    async def start_async_converter_from_object_to_file(cls, file_path: Path,
                                            sequence_of_objects: Iterable[OTLObject], **kwargs) -> None:

        await Helpers.converter_from_object_to_file(file_path=file_path,
                                              sequence_of_objects=sequence_of_objects,
                                              **kwargs)



    @classmethod
    @add_loading_screen
    async def converter_from_object_to_file(cls, file_path: Path, sequence_of_objects: Iterable[OTLObject], **kwargs) -> None:
        OTLLogger.logger.debug(f"Execute OtlmowConverter.from_objects_to_file({file_path.name})",
                               extra={"timing_ref": f"sequence_of_objects_{file_path.stem}"})
        await OtlmowConverter.from_objects_to_file_async(file_path=file_path,
                                             sequence_of_objects=sequence_of_objects,
                                             **kwargs)
        object_count = len(list(sequence_of_objects))
        OTLLogger.logger.debug(
            f"Execute OtlmowConverter.from_objects_to_file({file_path.name}) ({object_count} objects)",
            extra={"timing_ref": f"sequence_of_objects_{file_path.stem}"})

    @classmethod
    def is_version_equal_or_higher(cls, current_version:str, other_version:str) -> bool:
        res = Version(current_version) >= Version(other_version)
        OTLLogger.logger.info(f"Version comparison {current_version} >= {other_version} => {res}")
        return  res

    @classmethod
    def create_temp_path(cls, path_to_template_file_and_extension: Path) -> Path:
        """
        Creates a temporary path for storing files based on a template file.

        This method generates a temporary directory specifically for storing
        files related to the OTL project. It constructs the path using the
        name of the provided template file and ensures that the directory
        exists before returning the full path.

        :param cls: The class itself.
        :param path_to_template_file_and_extension: The path to the template file
                                                    for which the temporary path is created.
        :type path_to_template_file_and_extension: Path
        :returns: The path to the newly created temporary file location.
        :rtype: Path
        """

        tempdir = Helpers.get_base_temp_dir_path()
        if not tempdir.exists():
            os.makedirs(tempdir)
        doc_name = Path(path_to_template_file_and_extension).name
        return Path(tempdir) / doc_name

    @classmethod
    def get_base_temp_dir_path(cls):
        return Path(tempfile.gettempdir()) / 'temp-otlmow'

    @classmethod
    def extract_typeURI_from_aim_id(cls,aim_id:str,model_directory:Optional[Path]=None) -> Optional[str]:
        uuid = aim_id[:36]
        short_uri_encoded = aim_id[37:]

        if not validate_guid(uuid):
            return None
        if not short_uri_encoded:
            return None

        if missing_padding := len(short_uri_encoded) % 4:
            short_uri_encoded += '=' * (4 - missing_padding)

        short_uri_bytes = base64.b64decode(short_uri_encoded)
        try:
            short_uri = short_uri_bytes.decode('ascii')
        except UnicodeDecodeError:
            return None
        try:
            if short_uri == 'purl:Agent':
                agent = dynamic_create_instance_from_uri('http://purl.org/dc/terms/Agent',
                                                         model_directory=model_directory)
                return agent.typeURI

            ns, name = short_uri.split('#')
            instance:OTLObject = dynamic_create_instance_from_ns_and_name(ns, name,
                                                                          model_directory=model_directory)
            return instance.typeURI
        except ModuleNotFoundError:
            warnings.warn(
            'Could not import the module for the given aim_id, did you forget the model_directory?',
            category=ImportWarning)
            return None

        except ValueError:
            return None

    @classmethod
    def extract_corrected_relation_partner_typeURI(cls, partner_type_uri, partner_id,
                                                   id_to_typeURI_dict, ref_assets) -> Optional[
        str]:
        if partner_type_uri is None:
            if not id_to_typeURI_dict:
                id_to_typeURI_dict = {RelationChangeHelpers.get_corrected_identificator(asset): asset.typeURI
                                      for asset in ref_assets}

            if partner_id in id_to_typeURI_dict.keys():
                partner_type_uri = id_to_typeURI_dict[partner_id]

        # incase there is no asset with the correct assetId in the application
        if partner_type_uri is None and OTLObjectHelper.is_aim_id(partner_id):
            partner_type_uri = Helpers.extract_typeURI_from_aim_id(partner_id)

        return partner_type_uri

    @classmethod
    def open_folder_and_select_document(cls, document_path):
        document_path = str(document_path)
        if sys.platform.startswith("win"):
            command = ['explorer', '/select,', document_path]
            subprocess.Popen(command)
        elif sys.platform.startswith("darwin"):
            # macOS: open Finder and select the file
            command = ['open', '-R', document_path]
            subprocess.Popen(command)
        elif sys.platform.startswith("linux"):
            # Linux: open the containing folder with the default file manager
            folder = os.path.dirname(document_path)
            try:
                subprocess.Popen(['xdg-open', folder])
            except Exception as e:
                print(f"Could not open folder: {e}")
        else:
            print(f"Platform not supported for opening folder: {sys.platform}")

    @classmethod
    def compare_RelatieObjects(cls, obj1, obj2):
        dict1 = OTLObject.to_dict(obj1)
        dict2 = OTLObject.to_dict(obj2)

        if "isActief" in dict1:
            dict1.pop("isActief")
        if "isActief" in dict2:
            dict2.pop("isActief")

        return dict1 == dict2

    @classmethod
    def from_objects_to_file(cls, save_path,assets_in_memory):
        OtlmowConverter.from_objects_to_file(file_path=save_path,
                                             sequence_of_objects=assets_in_memory)

    @classmethod
    async def from_objects_to_file_async(cls, save_path,assets_in_memory):
        return await OtlmowConverter.from_objects_to_file_async(file_path=save_path,
                                                          sequence_of_objects=assets_in_memory)


import inspect
import os
from pathlib import Path
from typing import Optional, List, cast

from otlmow_model.OtlmowModel.BaseClasses.RelationInteractor import RelationInteractor
from otlmow_model.OtlmowModel.Classes.Agent import Agent
from otlmow_model.OtlmowModel.Classes.ImplementatieElement.AIMObject import AIMObject
from otlmow_model.OtlmowModel.Helpers import OTLObjectHelper
from otlmow_template.SubsetTemplateCreator import ROOT_DIR

from otlmow_gui.Domain import global_vars
from otlmow_gui.GUI.translation.GlobalTranslate import GlobalTranslate

ROOT_DIR_GUI = Path(__file__).parent.parent.parent.parent
SITE_PACKAGES_ROOT = ROOT_DIR
class RelationChangeHelpers:

    unspecified_direction_icon:str = "<->"
    outgoing_direction_icon: str = "-->"
    incoming_direction_icon: str = "<--"
    @classmethod
    def get_abbreviated_typeURI(cls, typeURI, add_namespace=False, is_relation=False):
        split_typeURI = typeURI.split("#")
        type_name = split_typeURI[-1]

        if typeURI == "http://purl.org/dc/terms/Agent":
            return typeURI.replace("http://purl.org/dc/terms/","")

        if is_relation:
            return typeURI.replace(
                "https://wegenenverkeer.data.vlaanderen.be/ns/onderdeel#", "")
        elif add_namespace:
            return typeURI.replace(
                "https://wegenenverkeer.data.vlaanderen.be/ns/", "")
        else:
            return type_name

    @classmethod
    def get_screen_name(cls, otl_object: RelationInteractor) -> Optional[str]:
        if otl_object is None:
            return None
        naam = cls.abbreviate_if_AIM_id(cls.get_corrected_identificator(otl_object))
        if otl_object.typeURI == 'http://purl.org/dc/terms/Agent':
            agent: Agent = cast(Agent, otl_object)
            # agent will always be external
            external_tranlation = GlobalTranslate._("external")

            if hasattr(agent, 'naam') and agent.naam:
                naam = " ".join([agent.naam,f"({external_tranlation})"])
            else:
                naam = " ".join([naam,f"({external_tranlation})"])
        else:
            aim_object: AIMObject = cast(AIMObject, otl_object)
            if hasattr(aim_object, 'naam') and aim_object.naam:
                naam = aim_object.naam
            else:
                naam = naam

            if aim_object.assetId.toegekendDoor == global_vars.external_toegekendDoor_label:
                external_tranlation = GlobalTranslate._("external")
                naam = " ".join([naam,f"({external_tranlation})"])

        return naam

    @classmethod
    def abbreviate_if_AIM_id(cls,id):
        return id.split("-")[0] + "-..." if OTLObjectHelper.is_aim_id(id) else id

    @classmethod
    def is_unique_across_namespaces(cls, typeURI,objects):
        split_typeURI = typeURI.split("#")
        type_name = split_typeURI[-1]

        unique_typeURIs = {otl_object.typeURI for otl_object in objects}
        list_type_names = [typeURI.split("#")[-1] for typeURI in unique_typeURIs]

        list_of_non_unique_type_names = filter(
            lambda type_name: list_type_names.count(type_name) < 2, list_type_names)

        return type_name not in list_of_non_unique_type_names

    @classmethod
    def get_screen_icon_direction(cls, input_richting:str) -> str:
        richting = cls.unspecified_direction_icon
        if input_richting == "Source -> Destination":
            richting = cls.outgoing_direction_icon
        elif input_richting == "Destination -> Source":
            richting = cls.incoming_direction_icon
        return richting


    @classmethod
    def list_all_non_abstract_class_type_uris(cls, otl_assets_only=False,
                                              model_directory='otlmow_model/OtlmowModel'):
        classes_to_instantiate = {}
        type_uri_list: List[str] = []

        class_location = Path(SITE_PACKAGES_ROOT) / model_directory / 'Classes/'

        for location, dirs, file in os.walk(class_location):
            dirs.remove("__pycache__")

            for dir in dirs:
                dir_location = Path(location, dir)

                for f in os.listdir(dir_location):
                    f = str(f)
                    if not os.path.isfile(dir_location / f):
                        continue
                    classes_to_instantiate[Path(f).stem] = Path(
                        dir_location / Path(f).stem).resolve()

        classes_to_instantiate['Agent'] = class_location / 'Agent'

        for class_name, file_path in classes_to_instantiate.items():
            
            import_path = f'{file_path.parts[-3]}.{file_path.parts[-2]}.{file_path.parts[-1]}'
            if "Agent" in str(file_path.absolute()):
                import_path = f'{file_path.parts[-2]}.{file_path.parts[-1]}'
            if 'otlmow_model' not in import_path:
                import_path = f'otlmow_model.OtlmowModel.{import_path}'

            try:
                py_mod = __import__(name=import_path, fromlist=f'{class_name}')
            except ModuleNotFoundError as e:
                raise ModuleNotFoundError(f'Could not import the module for {import_path}'
                ) from e

            class_ = getattr(py_mod, class_name)

            if not inspect.isabstract(class_):
                instance = class_()
                if hasattr(instance, "typeURI"):
                    if otl_assets_only and OTLObjectHelper.is_relation(instance):
                        continue
                    type_uri_list.append(instance.typeURI)
        return type_uri_list

    @classmethod
    def get_corrected_identificator(cls, otl_object: RelationInteractor):
        identificator = GlobalTranslate._("no_identificator")
        if hasattr(otl_object,"assetId"):
            identificator = str(otl_object.assetId.identificator)
        elif hasattr(otl_object,"agentId"):
            identificator = str(otl_object.agentId.identificator)

        return identificator



from otlmow_converter.Exceptions.UnknownExcelError import UnknownExcelError


class NoIdentificatorError(UnknownExcelError):


    def __init__(self,language, original_exception, tab: str = None):
        self._ = language
        super().__init__(original_exception, tab)

    def __str__(self):
        if self.tab:
            translation = self._("There are assets without an assetId.identificator in worksheet {0}")
            return translation.format(self.tab)
        else:
            return super().__str__()


import logging

from PyQt6.QtWidgets import QWidget, QMessageBox, QDialog, QVBoxLayout, QScrollArea, QLabel, \
    QPushButton, QApplication
import pyperclip as pc


class ErrorScreen(QWidget):

    def __init__(self, error_message, operation_continued=None):
        self.error_message = error_message
        pc.copy(error_message)
        super().__init__()
        self.operation_continued = operation_continued
        self.init_ui()

    def init_ui(self):
        # Create a QDialog for the error screen
        dlg = QDialog(self)
        dlg.setWindowTitle("Unhandled Exception")

        # Set up a layout
        layout = QVBoxLayout(dlg)

        # Create a scroll area
        scroll_area = QScrollArea(dlg)
        scroll_area.setWidgetResizable(True)



        # Add a QLabel inside the scroll area to display the error message
        content_widget = QWidget()
        content_layout = QVBoxLayout(content_widget)
        error_label = QLabel(self.error_message + "\n\nError message copied to clipboard.",
                             content_widget)
        error_label.setWordWrap(True)
        # if self.operation_continued:
        #     extra_message = QLabel(self.operation_continued)
        #     extra_message.setWordWrap(True)
        #     extra_message.setStyleSheet("color: green;")
        #     content_layout.addWidget(extra_message)

        content_layout.addWidget(error_label)
        scroll_area.setWidget(content_widget)

        # Scroll down to the bottom by default
        scroll_area.verticalScrollBar().rangeChanged.connect(
            lambda: scroll_area.verticalScrollBar().setValue(
                scroll_area.verticalScrollBar().maximum())
        )

        # Add a button to close the dialog
        close_button = QPushButton("Close", dlg)
        close_button.clicked.connect(dlg.close)

        # Add scroll area and button to the layout
        layout.addWidget(scroll_area)
        layout.addWidget(close_button)

        # Set the dialog size to half the screen size
        screen_geometry = QApplication.primaryScreen().availableGeometry()
        dlg.resize(int(screen_geometry.width()*(2/3)), int(screen_geometry.height()*(2/3)))

        # Show the dialog
        dlg.exec()


from pathlib import Path

from typing import Union, Callable
import datetime
import qtawesome as qta
from PyQt6.QtCore import Qt, QModelIndex
from PyQt6.QtGui import QPainter, QBrush, QColor

from PyQt6.QtWidgets import QTableWidget, QHeaderView, QStyledItemDelegate

from otlmow_gui.Domain.step_domain.HomeDomain import HomeDomain
from otlmow_gui.Domain.project.Project import Project
from otlmow_gui.Domain.util.Helpers import Helpers
from otlmow_gui.GUI.Styling import Styling
from otlmow_gui.GUI.dialog_windows.file_picker_dialog.ProjectExportFilePickerDialog import \
    ProjectExportFilePickerDialog
from otlmow_gui.GUI.dialog_windows.file_picker_dialog.SaveFilePickerDialog import SaveFilePickerDialog
from otlmow_gui.GUI.screens.Home_elements.OverviewTableItem import OverviewTableItem

from otlmow_gui.GUI.screens.general_elements.ButtonWidget import ButtonWidget

from otlmow_gui.GUI.dialog_windows.UpsertProjectWindow import UpsertProjectWindow
from otlmow_gui.GUI.dialog_windows.RemoveProjectWindow import RemoveProjectWindow


class LastAddedProjectHighlightDelegate(QStyledItemDelegate):

    def __init__(self, home_screen_widget):

        super().__init__()
        self.parent = home_screen_widget

    def paint(self, painter: QPainter, option, index: QModelIndex):
        painter.save()

        # Apply custom background for specific rows or items
        if self.parent.table.itemFromIndex(index.siblingAtColumn(0)).text() == self.parent.last_added_ref:
            painter.fillRect(option.rect, QBrush(Styling.last_added_color))

        painter.restore()
        super().paint(painter, option, index)

class OverviewTable(QTableWidget):

    def __init__(self, language_settings: Callable):
        super().__init__()
        self._ = language_settings
        self.projects: list
        self.main_window = None
        self.error_widget = OverviewTableItem()
        self.projects = None
        self.cellDoubleClicked.connect(self.open_project)
        self.export_project_dialog_window: SaveFilePickerDialog = (
            ProjectExportFilePickerDialog(self._))



    def draw_table(self) -> None:
        """
        Draws the table with specific configurations and settings.
        
        This method sets up the table by configuring its headers, enabling it, 
        and adjusting the visibility and behavior of the table elements. It 
        ensures that the table is not editable and that selection is done by 
        rows rather than individual cells.
        
        :param self: The instance of the class.
        
        :return: None
        """

        self.setEnabled(True)
        self.verticalHeader().setVisible(False)
        self.setColumnCount(7)
        self.setHorizontalHeaderLabels(
            [self._('own_reference'), self._('service_order'), self._('subset'), self._('last_edited'), '', '', ''])
        # ALign titles of header to the left
        self.horizontalHeader().setDefaultAlignment(Qt.AlignmentFlag.AlignLeft)
        for column in range(self.columnCount() - 2):
            self.horizontalHeader().setSectionResizeMode(column, QHeaderView.ResizeMode.Stretch)
        self.horizontalHeader().setSectionResizeMode(4, QHeaderView.ResizeMode.ResizeToContents)
        self.horizontalHeader().setSectionResizeMode(5, QHeaderView.ResizeMode.ResizeToContents)
        self.horizontalHeader().setSectionResizeMode(6, QHeaderView.ResizeMode.ResizeToContents)
        self.setShowGrid(False)
        # Zorgt ervoor dat selectie op row is niet op cell
        self.setSelectionBehavior(QTableWidget.SelectionBehavior.SelectRows)
        # Zorgt ervoor dat de table niet editable is
        self.setEditTriggers(QTableWidget.EditTrigger.NoEditTriggers)


    def fill_table(self, projects: [Project], search_text: str) -> None:
        """
        Fills the table with project data and applies a filter if necessary.
        
        This method populates the table with information from the provided 
        projects, adding the eigen referentie of the projects to each row as data. If a search text is 
        provided, it filters the displayed projects based on that text.
        
        :param self: The instance of the class.
        :param projects: A list of Project objects to be displayed in the table.
        :type projects: list[Project]
        :param search_text: A string used to filter the projects in the table.
        :type search_text: str
        
        :return: None
        """
        self.setSortingEnabled(False)
        self.setRowCount(len(projects))
        for row_index, project in enumerate(projects.values()):
            self.add_cell_to_table(row_index, 0, project.eigen_referentie)
            self.add_cell_to_table(row_index, 1, project.bestek)
            self.add_cell_to_table( row_index, 2,project.get_subset_db_name())
            self.add_cell_to_table( row_index, 3, project.laatst_bewerkt)
            self.add_action_buttons(row= row_index, project=project)
            self.add_data_to_row(row=row_index, eigen_referentie=project.eigen_referentie)

        if search_text:
            self.filter(search_text=search_text)
        self.setSortingEnabled(True)
        self.activate_initial_sort_on_last_edit()

    def add_cell_to_table(self, row: int, column: int, item: Union[str, datetime.datetime]) -> None:
        """Adds a cell to the table at the specified row and column with the given item.

           This method checks the type of the item being added to the table. If the item 
           is a date, it formats it as a string in the "dd-mm-yyyy" format; otherwise, 
           it converts the item to a string before adding it to the specified cell.

           :param self: The instance of the class.
           :param row: The row index where the cell will be added.
           :param column: The column index where the cell will be added.
           :param item: The item to be added to the cell, which can be a string or a datetime object.

           :return: None
           """
        if isinstance(item, datetime.date):
            list_item = OverviewTableItem(item.strftime(OverviewTableItem.date_format))
        else:
            list_item = OverviewTableItem(str(item))

        list_item.setToolTip(self._("Double-click to open project"))
        self.setItem(row, column,list_item)

    def add_action_buttons(self, row: int, project: Project) -> None:
        """
        Adds action buttons to the specified row for a given project.

        This method creates and configures three action buttons: an edit button, 
        a delete button, and a share button. Each button is connected to its respective 
        action, allowing users to edit project details, remove the project, or share 
        the project.

        :param self: The instance of the class.
        :param row: The row index where the buttons will be added.
        :type row: int
        :param project: The project object associated with the buttons.
        :type project: Project

        :return: None
        """
        edit_btn = ButtonWidget()
        edit_btn.setIcon(qta.icon('mdi.pencil'))
        edit_btn.setProperty('class','alter-button')
        edit_btn.clicked.connect(
            lambda _, input_projects=project: self.start_dialog_window(project=input_projects))
       

        delete_btn = ButtonWidget()
        delete_btn.setIcon(qta.icon('mdi.trash-can'))
        delete_btn.setProperty('class', 'alter-button')
        delete_btn.clicked.connect(lambda _, input_project=project:
                                   RemoveProjectWindow(language_settings=self._,
                                                       project=input_project))
       

        share_btn = ButtonWidget()
        share_btn.setIcon(qta.icon("mdi.share"))
        share_btn.setProperty('class', 'alter-button')
        share_btn.clicked.connect(lambda _, input_project=project:
                                  self.open_export_project_dialog_window(project=input_project))
        
        self.setCellWidget(row, 4, edit_btn)
        self.setCellWidget(row, 5, delete_btn)
        self.setCellWidget(row, 6, share_btn)

    def open_export_project_dialog_window(self, project:Project):

        # set the action_function in the dialog with the new project
        save_locations = self.export_project_dialog_window.summon(project_name=project.eigen_referentie)
        try:
            self.export_project(save_locations,project)
        except Exception as e:
            #TODO: proper error message when project fails to export
            raise e

    def export_project(self,save_locations: list[Path],project:Project):
        if not save_locations or not save_locations[0]:
            return
        project_path:Path = save_locations[0]
        if not project_path:
            return

        if project_path.suffix.lower().strip() != ('.otlw'):
            project_path = project_path.with_suffix('.otlw')

        project.export_project_to_file(file_path=project_path)
        Helpers.open_folder_and_select_document(project_path)

    def open_project(self, row) -> None:
        """
        Opens a project based on the specified row index.

        This method retrieves the project reference from the specified row in the 
        table and sets the main window to the project view. It then calls the 
        appropriate method to open the project using the retrieved reference.

        :param self: The instance of the class.
        :param row: The row index from which to retrieve the project reference.

        :return: None
        """

        # self.main_window.setCurrentIndex(1)
        project_ref = self.item(row, 0).data(1)

        HomeDomain.open_project(project_ref=project_ref)

    def filter(self, search_text: str = "") -> None:
        """
        Filters the table rows based on the absence of the provided search text.

        This method searches through all columns of the table for the specified
        search text. If the text is not found in any of the relevant fields,
        the corresponding rows are marked for removal from the table.

        Args:
            self: The instance of the class.
            search_text (str): The text to search for in the table rows. Defaults to an empty string.

        Returns:
            None
        """
        search_text = search_text.lower()
        rows_to_remove = []

        # search on all columns
        for i in range(self.rowCount()):
            eigen_referentie = self.item(i,0).text()
            bestek = self.item(i,1).text()
            subset = self.item(i,2).text()
            laatst_bewerkt = self.item(i,3).text()

            found = False
            highlight_brush = QBrush(QColor(Qt.GlobalColor.darkGray))
            if search_text in eigen_referentie.lower():
                self.item(i, 0).setBackground(highlight_brush)
                found = True
            if search_text in bestek.lower():
                self.item(i, 1).setBackground(highlight_brush)
                found = True
            if search_text in subset.lower():
                self.item(i, 2).setBackground(highlight_brush)
                found = True
            if search_text in laatst_bewerkt.lower():
                self.item(i, 3).setBackground(highlight_brush)
                found = True

            if not found:
                rows_to_remove.append(i)

        rows_to_remove.sort(reverse=True)

        for row_i in rows_to_remove:
            self.removeRow(row_i)

    def start_dialog_window(self, project: Project = None) -> None:
        """
        Starts a dialog window for upserting project information.

        This method initializes the UpsertProjectWindow, allowing users to
        create or update project details. It can optionally take a project
        object to pre-fill the dialog with existing project information.

        Args:
            self: The instance of the class.
            project (Project, optional): The project object to be edited. Defaults to None.

        Returns:
            None
        """

        window = UpsertProjectWindow(language_settings=self._,project=project)
        window.exec()

    def reset_ui(self, lang_settings: dict) -> None:
        """
        Resets the user interface with the provided language settings.

        This method updates the UI elements based on the specified language settings.
        It redraws the table, sets the horizontal header labels, and updates the error
        widget to display a message indicating no results.

        Args:
            self: The instance of the class.
            lang_settings (dict): A dictionary containing language settings for the UI.

        Returns:
            None
        """

        self._ = lang_settings
        self.draw_table()
        self.setHorizontalHeaderLabels(
            [self._('own_reference'), self._('service_order'), self._('subset'), self._('last_edited'), '', ''])
        self.error_widget.setText(self._('no_results'))

    def add_data_to_row(self, row: int, eigen_referentie: str) -> None:
        """
        Adds data to the specified row in the table.

        This method updates the first item in the specified row with the given
        eigen referentie value. It sets the data for the item at the first column
        of the specified row.

        Args:
            self: The instance of the class.
            row: The row index where the data will be added.
            eigen_referentie: The value to be set in the first item of the row.

        Returns:
            None
        """

        first_item_in_row = self.item(row, 0)
        first_item_in_row.setData(1,eigen_referentie)

    def activate_initial_sort_on_last_edit(self):
        self.sortItems(3,Qt.SortOrder.DescendingOrder)




from collections import namedtuple
from typing import Union

from PyQt6.QtCore import QItemSelectionModel, Qt
from PyQt6.QtGui import QStandardItem, QPixmap, QIcon, QFont
from PyQt6.QtWidgets import QFrame, QCheckBox
from otlmow_model.OtlmowModel.Helpers import OTLObjectHelper
from otlmow_model.OtlmowModel.BaseClasses.OTLObject import \
    OTLObject

from otlmow_gui.Domain.logger.OTLLogger import OTLLogger
from otlmow_gui.Domain.step_domain.RelationChangeDomain import RelationChangeDomain
from otlmow_gui.GUI.screens.RelationChange_elements.AbstractInstanceListWidget import \
    AbstractInstanceListWidget, IMG_DIR
from otlmow_gui.GUI.screens.RelationChange_elements.RelationChangeHelpers import RelationChangeHelpers
from otlmow_gui.exception_handler.ExceptionHandlers import create_task_reraise_exception


class PossibleRelationListWidget(AbstractInstanceListWidget):

    Text = namedtuple('text', ['typeURI', 'direction', 'screen_name', 'target_typeURI','full_typeURI'])
    Data = namedtuple('data', ['source_id', 'target_id', "index", "last_added"])

    def __init__(self, language_settings, parent):
        self._ = language_settings
        labels = [self._("related_asset"),self._("asset_type")]
        super().__init__(language_settings=language_settings,parent=parent, labels=labels, needs_source_object=True)

        self.list_label_text = self._('relations_list')
        self.list_subtext_label_text = self._("possible_relation_subscription")
        self.attribute_field_label_text = self._("possible_relation_partner_asset_attributes")
        self.show_all_OTL_relations_checkbox = QCheckBox()

    def create_object_list_gui(self, multi_select: bool = False) -> QFrame:
        frame = super().create_object_list_gui(multi_select)
        self.frame_layout.setContentsMargins(0,0,11,0)
        self.list_gui.setProperty('class', 'possible-relation-list')

        self.add_no_asset_selected_placeholder()

        return frame

    def add_no_asset_selected_placeholder(self):
        place_holder_item = QStandardItem(
            self._("Select_an_OTL-asset_to_see_possible_relations"))
        place_holder_item.setEditable(False)
        place_holder_item.setEnabled(False)
        place_holder_item.setSelectable(False)

        placeholder_font = QFont()
        placeholder_font.setItalic(True)
        place_holder_item.setFont(placeholder_font)

        # padding_item = QStandardItem("")
        # padding_item.setEditable(False)
        # padding_item.setEnabled(False)
        # padding_item.setSelectable(False)

        # self.list_gui.addItem([place_holder_item,padding_item ])
        self.list_gui.addItem([place_holder_item])
    def create_attribute_field(self):
        attribute_field = super().create_attribute_field()
        attribute_field.setProperty("class","attribute_field_possible_relation")
        return attribute_field

    def object_selected_listener(self, item) -> None:
        pass

    def on_item_selectionChange_listener(self, selected: QItemSelectionModel, deselected:QItemSelectionModel):
        no_item_selected = True

        for index in deselected.indexes():
            if index.column() == 0:
                item = self.list_gui.model.itemFromIndex(index)
                if item and item.isSelectable():
                    type_folder_item = item.parent()
                    self.reset_selected_item_count(type_folder_item=type_folder_item)

        dict_type_to_type_folder_item = {}
        dict_type_to_selected_item_count = {}
        # Get the currently selected indexes
        for index in self.list_gui.selectionModel().selectedIndexes():
            if index.column() == 0:
                item = self.list_gui.model.itemFromIndex(index)
                if item and item.isSelectable():
                    no_item_selected = False

                    # keep count of selected items in folder
                    parent_type_folder_item = item.parent()
                    parent_type_folder_type = parent_type_folder_item.data(self.data_1_index)
                    if parent_type_folder_type in dict_type_to_selected_item_count.keys():
                        dict_type_to_selected_item_count[parent_type_folder_type] += 1
                    else:
                        dict_type_to_selected_item_count[parent_type_folder_type] = 1
                        dict_type_to_type_folder_item[
                            parent_type_folder_type] = parent_type_folder_item

        # update the selected_counts on all type_folder_items
        for type_folder_type, selected_item_count in dict_type_to_selected_item_count.items():
            # OTLLogger.logger.debug( f"{type_folder_type}: {selected_item_count}")

            type_folder_item = dict_type_to_type_folder_item[type_folder_type]

            item_count = self.update_selected_count_data(type_folder_item=type_folder_item,
                                                         selected_item_count=selected_item_count)

            # apply new information to the folder_item display text
            self.set_type_folder_text(type_folder_item=type_folder_item,
                                      otl_type=type_folder_type,
                                      item_count=item_count,
                                      selected_item_count=selected_item_count)

        self.possible_relations_selected()
        self.set_list_button_enabled(not no_item_selected)

    def set_list_button_enabled(self, item_selected:bool):
        if not item_selected:
            if self.list_button.isEnabled():
                self.list_button.setEnabled(False)
        elif not self.list_button.isEnabled():
            self.list_button.setEnabled(True)

    def create_button(self):
        self.list_button.setText(self._('add_relation'))
        self.list_button.setDisabled(True)
        self.list_button.clicked.connect(
            self.add_possible_relation_to_existing_relations_listener)
        self.list_button.setProperty('class', 'add_relation-button')
        return self.list_button

    def create_instance_standard_item(self, text_and_data):
        text = f"{text_and_data['text'].screen_name}"
        instance_item = QStandardItem(f"{text}")
        instance_item.setData([text_and_data['data'].source_id,text_and_data['data'].target_id,text_and_data['data'].index], self.data_1_index)
        # instance_item.setData(text_and_data['data'].source_id, self.data_1_index)
        # instance_item.setData(text_and_data['data'].target_id, self.data_2_index)
        # instance_item.setData(text_and_data['data'].index, self.data_3_index)
        instance_item.setData("instance", self.item_type_data_index)
        instance_item.setData(text_and_data["data"].last_added, self.data_last_added_index)

        text2 = f"{text_and_data['text'].target_typeURI}"
        instance_item2 = QStandardItem(text2)

        pixmap = QPixmap(f'{str(IMG_DIR)}/bar_pipe.png')
        instance_item2.setIcon(QIcon(pixmap))

        self.add_direction_icon_to_item(instance_item=instance_item,
                                        direction=text_and_data['text'].direction,
                                        typeURI=text_and_data['text'].full_typeURI)

        return instance_item,instance_item2



    def add_possible_relation_to_existing_relations_listener(self):
        Data = self.Data
        data_list: list[Data] = sorted(self.get_selected_data(), reverse=True)

        create_task_reraise_exception(RelationChangeDomain.add_multiple_possible_relations_to_existing_relations(data_list=data_list))


        # for data in data_list:
        #     RelationChangeDomain.add_possible_relation_to_existing_relations(data.source_id,
        #                                                                      data.target_id,
        #                                                                      data.index)

    def possible_relations_selected(self):
        Data = self.Data # a named tuple type defined as variable of the class put into local var
        data_list: list[Data] = self.get_selected_data()

        create_task_reraise_exception(RelationChangeDomain.select_possible_relation_data(data_list))

    def get_selected_data(self):
        return [
            self.Data(self.list_gui.model.itemFromIndex(model_i).data(self.data_1_index)[0],
                      self.list_gui.model.itemFromIndex(model_i).data(self.data_1_index)[1],
                      self.list_gui.model.itemFromIndex(model_i).data(self.data_1_index)[2], False)

            for model_i in self.list_gui.selectionModel().selectedIndexes()
            if model_i.column() == 0] # we want one model_i per row, so only column == 0 is taken


    def extract_text_and_data_per_item(self, source_object: OTLObject, objects: Union[list[OTLObject],dict] , last_added):
        list_of_corresponding_values = []
        for target_identificator, target_relations in objects.items():

            for i, relation in enumerate(target_relations):
                relation_source_id:str = relation.bronAssetId.identificator
                relation_target_id: str = relation.doelAssetId.identificator

                target_object: OTLObject = RelationChangeDomain.get_object(
                                                                identificator=relation_target_id)
                try:
                    if target_object is None:
                        raise ValueError("target_object is None")
                    if source_object is None:
                        raise ValueError("source_object is None")
                except Exception as e:
                    OTLLogger.logger.debug(f"target or source are None in relation {relation.typeURI} \n{e}")
                    continue
                # if the target of the relation is the current selected object then you should
                # display the source object of the relation
                if target_object and source_object and target_object == source_object:
                    target_object =  RelationChangeDomain.get_object(
                                                          identificator=relation_source_id)

                #  if the new target_object is still the same as the source_object something is wrong
                if target_object and source_object and target_object == source_object:
                    raise ValueError()

                # determine the icon that indicates direction
                direction = ""
                if OTLObjectHelper.is_directional_relation(otl_object=relation):
                    if relation_source_id == target_identificator:
                        direction = RelationChangeHelpers.incoming_direction_icon

                    else:
                        direction = RelationChangeHelpers.outgoing_direction_icon
                else:
                    direction = RelationChangeHelpers.unspecified_direction_icon

                real_source_id: str = RelationChangeHelpers.get_corrected_identificator(
                    otl_object=source_object)
                abbr_relation_typeURI: str = RelationChangeHelpers.get_abbreviated_typeURI(
                    typeURI=relation.typeURI,
                    add_namespace=False,
                    is_relation=OTLObjectHelper.is_relation(relation))
                target_screen_name:str = RelationChangeHelpers.get_screen_name(otl_object=target_object)

                try:
                    add_target_namespace:bool = RelationChangeHelpers.is_unique_across_namespaces(
                        typeURI=target_object.typeURI,
                        objects=RelationChangeDomain.shown_objects)
                    abbr_target_object_typeURI:str = RelationChangeHelpers.get_abbreviated_typeURI(
                        typeURI=target_object.typeURI,
                        add_namespace=add_target_namespace,
                        is_relation=OTLObjectHelper.is_relation(target_object))

                    is_last_added:bool =    (relation.assetId.identificator in
                                            [e.assetId.identificator for e in last_added])

                    list_of_corresponding_values.append({
                        "text": self.Text(typeURI=abbr_relation_typeURI,
                                          direction=direction,
                                          screen_name=target_screen_name,
                                          target_typeURI=abbr_target_object_typeURI,
                                          full_typeURI=relation.typeURI),
                        "data": self.Data(source_id=real_source_id ,
                                          target_id=target_identificator,
                                          index=i,
                                          last_added=is_last_added)
                    })
                except Exception as e:
                    target_screen_name: str = RelationChangeHelpers.get_screen_name(
                        otl_object=target_object)
                    abbr_relation_typeURI: str = RelationChangeHelpers.get_abbreviated_typeURI(
                        typeURI=relation.typeURI,
                        add_namespace=False,
                        is_relation=OTLObjectHelper.is_relation(relation))
                    real_source_id: str = RelationChangeHelpers.get_corrected_identificator(
                        otl_object=source_object)
                    OTLLogger.logger.debug(f"Couldn't make relation {abbr_relation_typeURI}: {real_source_id} {direction} {target_screen_name} because \n{e}")
        list_of_corresponding_values.sort(key=lambda val: (
            val['text'].target_typeURI, val['text'].screen_name, val['text'].typeURI))
        return list_of_corresponding_values

    def is_last_added(self, text_and_data: dict):
        return text_and_data["data"].last_added

    def create_asset_type_standard_item(self, asset_type:str, text_and_data_list) -> QStandardItem:
        folder_item = super().create_asset_type_standard_item(asset_type,text_and_data_list)

        if text_and_data_list:
            full_typeURI = text_and_data_list[0]["text"].full_typeURI
            self.add_colored_relation_bol_icon_to_item(folder_item, full_typeURI)
        return folder_item

    def get_no_instance_selected_message(self) -> str:
        return self._("no_relation_selected")

    def add_extra_elements_to_list_subtext_layout(self) -> None:
        self.show_all_OTL_relations_checkbox.setText(self._("Toon all OTL-relaties"))
        self.show_all_OTL_relations_checkbox.setToolTip(self._("Toon alle mogelijke relaties met het geselecteerde asset in het volledige OTL (i.p.v. alleen subset)"))
        self.show_all_OTL_relations_checkbox.stateChanged.connect(self.state_change_OTL_relations_checkbox)

        self.list_subtext_layout.addWidget(self.show_all_OTL_relations_checkbox)


    def is_show_all_OTL_relations_checked(self) -> bool:
        return self.show_all_OTL_relations_checkbox.isChecked()

    def state_change_OTL_relations_checkbox(self,state:int) -> None:

        if state == Qt.CheckState.Unchecked.value:
            RelationChangeDomain.set_search_full_OTL_mode(state=False)
        elif state == Qt.CheckState.Checked.value:
            RelationChangeDomain.set_search_full_OTL_mode(state=True)
        else:
            # In case of an unexpected state, default to False.
            RelationChangeDomain.set_search_full_OTL_mode(state=False)

        RelationChangeDomain.update_frontend()

    def update_this_gui_list_content(self):
        RelationChangeDomain.update_frontend_possible_relations()





import os
import sys
from pathlib import Path
from typing import Optional

test_mode = False

projects = []
current_project = None
otl_wizard = None


# SDF files are only supported in Windows systems
if "win" in sys.platform:
    supported_file_formats = {'Excel':"xlsx", 'CSV':"csv", 'JSON':'json','GeoJSON': 'geojson','SDF':'sdf'}
else:
    supported_file_formats = {'Excel': "xlsx", 'CSV': "csv", 'JSON': 'json', 'GeoJSON': 'geojson'}

external_toegekendDoor_label = "OTL_wizard_2"
last_subset_selected_dir: Optional[Path] = None
FDO_toolbox_path_str =  'C:\\Program Files\\FDO Toolbox\\FdoCmd.exe'
FDO_toolbox_installer_path_str =  str(Path(os.getcwd()) / "additional_programs" / "FDOToolbox-Release-v1.5.3-x64-Setup.exe")

def get_start_dir_subset_selection(input_subset_str:str) -> Optional[str]:
    if input_subset_str:
        return os.path.dirname(input_subset_str)
    elif last_subset_selected_dir:
        return last_subset_selected_dir
    else:
        return None



from typing import Optional

from otlmow_gui.GUI.dialog_windows.file_picker_dialog.LoadFilePickerDialog import LoadFilePickerDialog


class ProjectImportFilePickerDialog(LoadFilePickerDialog):


    def __init__(self, language_settings, title: Optional[str] = None,
                 name_filter: str = None, initial_file_path: Optional[str] = None):
        
        super().__init__(language_settings, title, name_filter, initial_file_path)
        if not title:
            self.setWindowTitle(self._("select_OTL_project"))
        if not name_filter:
            self.setNameFilter(self._("OTLWizard project files (*.otlw)"))


import datetime
import gettext
import json
import logging
import platform
from pathlib import Path

from otlmow_gui.Domain.enums import Language
from otlmow_gui.Domain.logger.OTLLogger import OTLLogger
from otlmow_gui.Domain.ProgramFileStructure import ProgramFileStructure


class Settings:
    settings_filename = 'settings.json'

    @classmethod
    def return_language(cls, locale_dir: Path, language: Language = Language.DUTCH):
        """
        Sets the application's language by loading the appropriate translation files.

        This class method retrieves the translation for the specified language from the given
        locale directory, defaulting to Dutch if no language is provided.

        :param locale_dir: The directory where the translation files are located.
        :type locale_dir: Path
        :param language: The language to be set for translations. If not provided, defaults to
        Language.DUTCH.
        :type language: Language, optional

        :return: A gettext function that can be used to translate messages.
        :rtype: function

        :raises Exception: If there is an error loading the translation files.

        :example:
            translator = Settings.return_language(Path('/path/to/locales'), Language.ENGLISH)

        Falls back to returning the original string if no translation is found.
        """
        language_str = str(language)
        logging.debug(f"Changing language to: {language_str}")

        try:
            translator = gettext.translation(
                'messages',
                localedir=locale_dir,
                languages=[language.value]
            )
            translator.install()
            return translator.gettext
        except FileNotFoundError:
            OTLLogger.logger.warning(
                f"Translation file for '{language.value}' not found in {locale_dir}. "
                "Falling back to identity translation."
            )
            return lambda x: x
        except Exception as e:
            OTLLogger.logger.error(f"Error loading translations for {language.value}: {e}")
            return lambda x: x

    @classmethod
    def get_or_create_settings_file(cls) -> dict:
        """
        Retrieves or creates a settings file for the application.

        This class method checks for the existence of a settings file and initializes it with
        default values if it does not exist, including the operating system, language, and run
        timestamps.

        :return: A dictionary containing the settings details, including language, operating system, first run status, and last run timestamp.
        :rtype: dict

        :raises IOError: If there is an error reading or writing the settings file.

        :example:
            settings = Settings.get_or_create_settings_file()
        """


        work_dir_path = ProgramFileStructure.get_otl_wizard_work_dir()
        settings_filepath = work_dir_path / 'settings.json'
        first_run = not settings_filepath.exists()
        timestamp = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")

        operating_sys = platform.system()
        language = Language.DUTCH


        settings_details = cls.load_settings(settings_filepath)

        if not settings_details.__contains__('language'):
            settings_details['language'] = str(language.name)

        settings_details['OS'] = str(operating_sys)
        settings_details['first_run'] = first_run
        settings_details['last_run'] = timestamp

        cls.save_settings(settings_details, settings_filepath)

        return settings_details

    @classmethod
    def save_settings(cls, settings_details, settings_filepath):
        with open(settings_filepath, 'w') as json_file:
            json.dump(settings_details, json_file)

    @classmethod
    def change_language_on_settings_file(cls, lang) -> None:
        """
        Updates the language setting in the application's settings file.

        This class method reads the current settings, modifies the language entry to the
        specified value, and saves the updated settings back to the file.

        :param cls: The class that calls this method.
        :type cls: type
        :param lang: The new language to be set, represented as a Language enum.
        :type lang: Language

        :return: None

        :raises FileNotFoundError: If the settings file does not exist.
        :raises json.JSONDecodeError: If there is an error reading the settings file.
        :raises IOError: If there is an error writing to the settings file.

        :example:
            Settings.change_language_on_settings_file(Language.ENGLISH)
        """

        work_dir_path = ProgramFileStructure.get_otl_wizard_work_dir()
        settings_file = work_dir_path / cls.settings_filename
        with open(settings_file) as json_file:
            settings_details = json.load(json_file)
        settings_details['language'] = str(lang.name)
        with open(settings_file, 'w') as f:
            json.dump(settings_details, f)

    @classmethod
    def get_language_from_settings(cls) -> Language:
        """
        Retrieves the current language setting from the application's settings file.

        This class method reads the settings file and returns the language specified within it as
        a Language enum.

        :return: The language currently set in the settings file.
        :rtype: Language

        :raises FileNotFoundError: If the settings file does not exist.
        :raises json.JSONDecodeError: If there is an error reading the settings file.
        :raises KeyError: If the 'language' key is not found in the settings details.

        :example:
            current_language = Settings.get_language_from_settings()
        """

        work_dir_path = ProgramFileStructure.get_otl_wizard_work_dir()
        settings_file = work_dir_path / cls.settings_filename
        with open(settings_file) as json_file:
            settings_details = json.load(json_file)
        return Language[settings_details['language']]

    @classmethod
    def load_settings(cls,settings_filepath:Path):
        try:
            with open(settings_filepath, 'r') as json_file:
                return json.load(json_file)
        except (FileNotFoundError, json.JSONDecodeError):
            return {}


from collections import namedtuple

from PyQt6.QtCore import QItemSelectionModel
from PyQt6.QtGui import QStandardItem, QFont
from PyQt6.QtWidgets import QFrame
from otlmow_model.OtlmowModel.Helpers import OTLObjectHelper

from otlmow_gui.Domain.step_domain.RelationChangeDomain import RelationChangeDomain
from otlmow_gui.GUI.dialog_windows.AddExternalAssetWindow import AddExternalAssetWindow
from otlmow_gui.GUI.screens.RelationChange_elements.AbstractInstanceListWidget import \
    AbstractInstanceListWidget

from otlmow_gui.GUI.screens.RelationChange_elements.RelationChangeHelpers import RelationChangeHelpers
from otlmow_gui.exception_handler.ExceptionHandlers import create_task_reraise_exception


class ObjectListWidget(AbstractInstanceListWidget):
    Text = namedtuple('text', ['typeURI', 'screen_name', 'full_typeURI'])
    Data = namedtuple('data', ['selected_object_id','last_added'])

    def __init__(self, language_settings,parent):
        self._ = language_settings
        labels = [ self._("OTL_asset")]
        super().__init__(language_settings=language_settings,parent=parent,labels=labels)

        self.list_label_text = self._('class_list')
        self.list_subtext_label_text = self._("otl_asset_description")
        self.attribute_field_label_text = self._("object_attributes")

    def create_object_list_gui(self, multi_select: bool = False) -> QFrame:
        frame = super().create_object_list_gui(multi_select)
        self.frame_layout.setContentsMargins(11,0,0,0)
        self.list_gui.setProperty('class', 'object-list')
        return frame

    def on_item_selectionChange_listener(self, selected: QItemSelectionModel, deselected:QItemSelectionModel):
        # sourcery skip: remove-dict-keys

        if self.selected_item:
            type_folder_item = self.selected_item.parent()
            self.reset_selected_item_count(type_folder_item=type_folder_item)

        # Get the currently selected indexes
        self.selected_object = None
        dict_type_to_type_folder_item = {}
        dict_type_to_selected_item_count = {}
        for index in self.list_gui.selectionModel().selectedIndexes():
            if index.column() == 0:
                item = self.list_gui.model.itemFromIndex(index)
                if item and item.isSelectable():
                    selected_object_id = item.data(self.data_1_index)

                    self.selected_object = RelationChangeDomain.get_object(
                        identificator=selected_object_id)
                    self.selected_item = item



                    # keep count of selected items in folder
                    parent_type_folder_item = item.parent()
                    parent_type_folder_type = parent_type_folder_item.data(self.data_1_index)
                    if parent_type_folder_type in dict_type_to_selected_item_count.keys():
                        dict_type_to_selected_item_count[parent_type_folder_type] += 1
                    else:
                        dict_type_to_selected_item_count[parent_type_folder_type] = 1
                        dict_type_to_type_folder_item[parent_type_folder_type] = parent_type_folder_item

        # update the selected_counts on all type_folder_items
        for type_folder_type, selected_item_count in dict_type_to_selected_item_count.items():
            # OTLLogger.logger.debug( f"{type_folder_type}: {selected_item_count}")

            type_folder_item = dict_type_to_type_folder_item[type_folder_type]

            item_count = self.update_selected_count_data(type_folder_item=type_folder_item,
                                                         selected_item_count=selected_item_count)

            # apply new information to the folder_item display text
            self.set_type_folder_text(type_folder_item=type_folder_item,
                                      otl_type=type_folder_type,
                                      item_count=item_count,
                                      selected_item_count=selected_item_count)

        create_task_reraise_exception(RelationChangeDomain.set_possible_relations(selected_object=self.selected_object))

    def select_item_via_identificator(self,identificator):
        for i in range(self.list_gui.model.rowCount()):

            folder_item = self.list_gui.model.item(i)
            for j in range(folder_item.rowCount()):
                item = folder_item.child(j)
                if item.data(self.data_1_index) == identificator:
                    self.select_object_id(item)
                    self.parent.set_existing_relation_search_bar_text(item.text())

    def create_button(self):
        self.list_button.setEnabled(True)
        self.list_button.setText(self._("add_external_asset"))
        self.list_button.clicked.connect(
            self.add_external_asset)
        self.list_button.setProperty("class", 'add-external-button')
        return self.list_button

    def add_external_asset(self):
        add_asset_window = AddExternalAssetWindow(self._)
        add_asset_window.draw_add_external_asset_window()

    def is_previously_selected_requirement(self, text_and_data):
        return self.selected_object and RelationChangeHelpers.get_corrected_identificator(self.selected_object) == text_and_data['data'].selected_object_id

    def extract_text_and_data_per_item(self, source_object, objects, last_added):
        list_of_corresponding_values = []
        self.id_to_object_with_text_and_data_dict.clear() # for usage in MapScreen

        for OTL_object in objects:
            screen_name = RelationChangeHelpers.get_screen_name(OTL_object)
            add_namespace = RelationChangeHelpers.is_unique_across_namespaces(
                OTL_object.typeURI,
                RelationChangeDomain.shown_objects)
            abbr_typeURI = RelationChangeHelpers.get_abbreviated_typeURI(
                OTL_object.typeURI,
                add_namespace,
                OTLObjectHelper.is_relation(OTL_object))

            correct_id:str = RelationChangeHelpers.get_corrected_identificator(OTL_object)
            text_and_data = {
                "text": self.Text(abbr_typeURI,screen_name,OTL_object.typeURI),
                "data": self.Data(correct_id, False)
            }
            list_of_corresponding_values.append(text_and_data)

            self.id_to_object_with_text_and_data_dict[correct_id] = [OTL_object,text_and_data]
        return list_of_corresponding_values
    def create_instance_standard_item(self, text_and_data):
        text = f"{text_and_data['text'].screen_name}"
        instance_item = QStandardItem(text)
        instance_item.setData(text_and_data['data'].selected_object_id, self.data_1_index)
        instance_item.setData("instance", self.item_type_data_index)
        instance_item.setData(text_and_data["data"].last_added, self.data_last_added_index)

        return instance_item,

    def select_object_id(self, previously_selected_item: QStandardItem):
        if previously_selected_item:
            previously_selected_item_index = self.list_gui.model.indexFromItem(
                previously_selected_item)
            if previously_selected_item_index:
                # renew selected_item before new selection
                self.selected_item = previously_selected_item

                self.list_gui.selectionModel().setCurrentIndex(
                    previously_selected_item_index,
                    QItemSelectionModel.SelectionFlag.SelectCurrent)

    def is_last_added(self, text_and_data: dict):
        pass

    def add_no_options_placeholder(self):
        place_holder_item = QStandardItem(
            self._("no_options_available"))
        place_holder_item.setEditable(False)
        place_holder_item.setEnabled(False)
        place_holder_item.setSelectable(False)

        placeholder_font = QFont()
        placeholder_font.setItalic(True)
        place_holder_item.setFont(placeholder_font)

        self.list_gui.addItem(place_holder_item)

    def asset_clicked_listener(self):
        # automatically set the 3rd column (existing relations) to search for
        # the selected asset
        if self.selected_item:
            self.parent.set_existing_relation_search_bar_text(self.selected_item.text())

            # highlight on map
            if self.parent.map_window:

                selected_object_id = self.selected_item.data(self.data_1_index)
                self.selected_object = RelationChangeDomain.get_object(
                    identificator=selected_object_id)
                selected_id = RelationChangeHelpers.get_corrected_identificator(
                    otl_object=self.selected_object)

                if selected_id:
                    self.parent.map_window.activate_highlight_layer_by_id(selected_id)

    def update_this_gui_list_content(self):
        RelationChangeDomain.update_frontend_objects()




import logging
import os
import re
import subprocess
import sys
from pathlib import Path
from typing import Optional

from otlmow_gui.Domain import global_vars
from otlmow_gui.Domain.util.Helpers import Helpers
from otlmow_gui.Domain.util.XSDCreator import XSDCreator
from otlmow_gui.Domain.logger.OTLLogger import OTLLogger
from otlmow_gui.Exceptions.FDOToolboxNotInstalledError import FDOToolboxNotInstalledError

ROOT_DIR =  Path(Path(__file__).absolute()).parent.parent
sys.path.insert(0,str(ROOT_DIR.absolute()))# needed for python to import project files

from otlmow_gui.Exceptions.FDOToolboxProcessError import FDOToolboxProcessError
from otlmow_gui.Exceptions.WrongFileTypeError import WrongFileTypeError
from otlmow_gui.GUI.translation.GlobalTranslate import GlobalTranslate


class SDFHandler:
    """
    documentation for FDOcmd.exe:
    https://jumpinjackie.github.io/fdotoolbox/userdoc/1.5.0/cmdline.html?highlight=fdocmd
    """


    @classmethod
    def _get_classes_from_SDF_file(cls, sdf_file_path:Path) -> list[str]:

        sdf_file_path_str = sdf_file_path.absolute()
        command = f'"{global_vars.FDO_toolbox_path_str}" list-classes --from-file "{sdf_file_path_str}"'

        output, error = cls.run_command(command=command)

        if error:
            cls._filter_out_coordinate_system_not_installed_error(command=command, error=error)

        if output:
            return output.split("\n")
        else:
            return []

    @classmethod
    def _filter_out_coordinate_system_not_installed_error(cls, command:str, error:str) -> None:
        # The following error occurs even though the program works properly
        # Here we filter out this error to be able to capture and use other critical errors
        pattern1 = re.compile(
            r'\(\d+\) DefaultDir: "C:\\ProgramData\\Autodesk\\Geospatial Coordinate Systems" is not a directory! Install the Coordinate System library into this directory or set MENTOR_DICTIONARY_PATH to where they are currently installed\.\n'
            r'\(\d+\) MgCoordinateSystemInitializationFailedException caught in CCoordinateSystemCatalog constructor - The coordinate system initialization failed\.\n\n'
            r'- MgCoordinateSystemCatalog\.SetDefaultDictionaryDirAndFileNames\(\) line 603 file \.\.\\CoordinateSystem\\CoordSysCatalog\.cpp\n'
            r'- MgCoordinateSystemCatalog\.GetDefaultDictionaryDir\(\) line 263 file \.\.\\CoordinateSystem\\CoordSysCatalog\.cpp'
        )
        filtered_error, instance_count = pattern1.subn("", error)

        # pattern2 = re.compile()
        # The following error occurs with the create-file command even though the program works properly
        # Here we filter out this error to be able to capture and use other critical errors
        # negative_precision_error = ('\n'
        #                      '\n'
        #                      'OSGeo.FDO.Common.Exception: A Data Property cannot have a negative '
        #                      'precision. \n'
        #                      '   at OSGeo.FDO.Schema.DataPropertyDefinition.set_Precision(Int32 value)\n'
        #                      '   at '
        #                      'FdoToolbox.Core.Feature.SchemaCapabilityChecker.FixDataProperties(ClassDefinition& '
        #                      'classDef)\n'
        #                      '   at '
        #                      'FdoToolbox.Core.Feature.SchemaCapabilityChecker.AlterClassDefinition(ClassDefinition '
        #                      'classDef, IncompatibleClass incClass, Func`1 getActiveSpatialContext, '
        #                      'Action`2 fixGeomSc)\n'
        #                      '   at '
        #                      'FdoToolbox.Core.Feature.SchemaCapabilityChecker.AlterSchema(FeatureSchema '
        #                      'schema, IncompatibleSchema incompatibleSchema, Func`1 '
        #                      'getActiveSpatialContext)\n'
        #                      '   at FdoToolbox.Core.Feature.FdoFeatureService.LoadSchemasFromXml(String '
        #                      'xmlFile, Boolean fix)\n'
        #                      '   at FdoCmd.Commands.CreateFileCommand.Execute()')
        #
        # filtered_error = filtered_error.replace(negative_precision_error,"")

        if filtered_error:
            raise FDOToolboxProcessError(language=GlobalTranslate._, command=command,
                                         fdo_toolbox_error=filtered_error)


    @classmethod
    def _validate_SDF_file(cls, sdf_filepath) -> None:
        if not sdf_filepath.exists():
            raise FileNotFoundError(f'{sdf_filepath} is not a valid path. File does not exist.')
        if sdf_filepath.suffix != ".sdf":
            raise WrongFileTypeError(language=GlobalTranslate._,
                                     expected_filetype_name="SDF-file",
                                     expected_filetype_suffix=".sdf")

    @classmethod
    def _get_objects_from_class(cls, sdf_filepath, sdf_class) -> str:

        sdf_file_path_str = sdf_filepath.absolute()
        command = f'"{global_vars.FDO_toolbox_path_str}" query-features --class "{sdf_class}" --from-file "{sdf_file_path_str}"  --format CSV'

        output, error = cls.run_command(command)
        if error:
            cls._filter_out_coordinate_system_not_installed_error(command, error)

        output = output.replace("_",".")
        output = output.replace("Geometry", "geometry")
        output = output.replace("XYZ", "Z")

        return output + "\n" # added for compatibility with old OTL-wizard csv generation

    @classmethod
    def convert_SDF_to_CSV(cls, sdf_filepath:Path, csv_output_path:Path) -> list[str]:

        OTLLogger.logger.debug(
            f"Executing Domain.SDFHandler.SDFHandler.convert_SDF_to_CSV for {sdf_filepath.name}",
            extra={"timing_ref": f"validate_{sdf_filepath.name}"})

        output_csv_filepath_list = []

        if not cls.check_if_FDOToolbox_is_installed():
            pass
            # return []
        cls._validate_SDF_file(sdf_filepath)

        # format the expected output csv files based on the classes in the sdf file
        output_classes: list[str] = SDFHandler._get_classes_from_SDF_file(sdf_filepath)
        if ((os.path.exists(csv_output_path) and os.path.isdir(csv_output_path)) or
            csv_output_path.suffix == ""):
            csv_output_path_is_dir = True
            output_basepath: str = str(csv_output_path)
            os.makedirs(csv_output_path, exist_ok=True)
        else:
            csv_output_path_is_dir = False
            output_basepath: str = os.path.splitext(csv_output_path)[0]
            os.makedirs(csv_output_path.parent, exist_ok=True)

        for otlclass in output_classes:

            objects_str:str = SDFHandler._get_objects_from_class(sdf_filepath=sdf_filepath, sdf_class=otlclass)



            # build absolute path to csv of output
            if csv_output_path_is_dir:
                filepath_of_output_csv_for_one_class = str(
                    Path(output_basepath) / f"{otlclass}.csv")
            else:
                filepath_of_output_csv_for_one_class = output_basepath + otlclass + ".csv"

            with open(filepath_of_output_csv_for_one_class,mode='w+') as output_csv_file:
                output_csv_file.write(objects_str)

            logging.debug(f"created file: {filepath_of_output_csv_for_one_class}")
            output_csv_filepath_list.append(filepath_of_output_csv_for_one_class)

        OTLLogger.logger.debug(
            f"Executing Domain.SDFHandler.SDFHandler.convert_SDF_to_CSV for {sdf_filepath.name}",
            extra={"timing_ref": f"validate_{sdf_filepath.name}"})

        return output_csv_filepath_list

    @classmethod
    def run_command(cls,command):
        # Using subprocess.run to execute the command without opening a terminal window
        result = subprocess.run(command, shell=True, stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE, text=True)
        return result.stdout.strip(), result.stderr.strip()

    @classmethod
    def check_if_FDOToolbox_is_installed(cls) -> bool:

        try:
            if not os.path.exists(global_vars.FDO_toolbox_path_str):
                raise FDOToolboxNotInstalledError(GlobalTranslate._)
        except FDOToolboxNotInstalledError as e:
            #only show the notification window if the GUI MainWindow object exists²
            if global_vars.otl_wizard and global_vars.otl_wizard.main_window:


                answer = global_vars.otl_wizard.main_window.show_blocking_yes_no_notification_window(
                    text=str(e),
                    title=GlobalTranslate._("FDOToolbox not installed")
                )

                if answer != 16384: # QMessageBox.ButtonRole.NoRole:
                    return False

                test_path = Path(os.getcwd()) / Path(
                    "LatestReleaseMulti\\additional_programs\\FDOToolbox-Release-v1.5.3-x64-Setup.exe")

                if os.path.exists(global_vars.FDO_toolbox_installer_path_str):
                    OTLLogger.logger.debug(f"go to {global_vars.FDO_toolbox_installer_path_str}")
                    subprocess.Popen(f'explorer /select,"{global_vars.FDO_toolbox_installer_path_str}"')
                else:
                    OTLLogger.logger.debug(f"go to {test_path}")
                    subprocess.Popen(
                        f'explorer /select,"{test_path}"')

                return False

                # raise FDOToolboxNotInstalledError(GlobalTranslate._) from e
        return True

    @classmethod
    def _convert_XSD_to_SDF(cls, input_xsd_path:Path, output_sdf_path:Path) -> None:
        sdf_file_path_str = output_sdf_path.absolute()
        input_xsd_path_str = input_xsd_path.absolute()

        command = (f'"{global_vars.FDO_toolbox_path_str}" create-file '
                   f'--file "{sdf_file_path_str}"  --schema-path "{input_xsd_path_str}"')

        output, error = cls.run_command(command)
        OTLLogger.logger.debug(f"convert_XSD_to_SDF:\n{output}")

        if error:

            cls._filter_out_coordinate_system_not_installed_error(command, error)

    @classmethod
    async def create_filtered_SDF_from_subset(cls, subset_path: Path, sdf_path: Path,
                                        selected_classes_typeURI_list: Optional[list[str]]=None,
                                        model_directory: Path = None) -> None:

        if not cls.check_if_FDOToolbox_is_installed():
            pass

        temp_path: Path = Helpers.create_temp_path(path_to_template_file_and_extension=sdf_path)
        temp_path = temp_path.parent / f'{temp_path.name}.xsd'

        await XSDCreator.create_filtered_xsd_from_subset(
            subset_path=subset_path,xsd_path=temp_path,
            selected_classes_typeURI_list=selected_classes_typeURI_list,
            model_directory=model_directory)

        SDFHandler._convert_XSD_to_SDF(input_xsd_path=temp_path,output_sdf_path=sdf_path)

if __name__ == "__main__":
    logger = logging.getLogger()
    logger.setLevel(logging.DEBUG)
    settings = {"language": "ENGLISH"}
    root = Path(Path(__file__).absolute()).parent.parent
    GlobalTranslate(settings, root/ 'locale')
    # sdf_path_input = Path("C:\\Users\\chris\\Documents\\job_related\\wegen_en_verkeer\\new_python_otl_wizard\\testData\\Ruben_SDF_test\\DA-2024-03992_export\\download.sdf")
    sdf_path_input = Path(
        "C:\\Users\\chris\\Documents\\job_related\\wegen_en_verkeer\\new_python_otl_wizard\\testData\\Ruben_SDF_test\\DA-2025-00023_export\\download.sdf")

    # sdf_file_classes = SDFHandler.get_classes_from_SDF_file(sdf_file_path=sdf_path_input)
    # sdf_objects = SDFHandler._get_objects_from_class(sdf_file_path=sdf_path_input,sdf_class="OTL_Brandblusser")
    sdf_objects = SDFHandler._get_objects_from_class(sdf_filepath=sdf_path_input, sdf_class="OTL_Voedingskabel")
    print(sdf_objects)

    # write output to test files
    # output = root / "UnitTests\\test_files\\output_ref\\output_get_object_from_class_DA-2025-00023_export.txt"
    # output_csv = root / "UnitTests\\test_files\\output_test\\convert_SDF_to_CSV_DA-2025-00023_export\\test_DA-2024-03992_export.csv"
    # output_csv = root / "C:\\Users\\chris\\Documents\\job_related\\wegen_en_verkeer\\new_python_otl_wizard\\testData\\Ruben_SDF_test\\DA-2025-00023_export\\christiaan_omzetting"
    # SDFHandler.convert_SDF_to_CSV(sdf_path_input,output_csv)




from otlmow_converter.Exceptions.UnknownExcelError import UnknownExcelError


class RelationHasNonExistingTypeUriForSourceOrTarget(UnknownExcelError):
    """
    The bronAssetID.typeURI or doelAssetID.typeURI do NOT EXIST as real TypeUri's
    This precludes the fact that these typeURI's are invalid for the current relationTypeURI
    """
    def __init__(self,relation_type_uri,relation_identificator,wrong_field,wrong_value,tab=None,*args):
        super().__init__(Exception(),*args)
        self.relation_type_uri = relation_type_uri
        self.relation_identificator = relation_identificator
        self.wrong_field = wrong_field
        self.wrong_value = wrong_value
        self.tab = tab